import abc
from _typeshed import Incomplete
from abc import abstractmethod
import typing as T
from types import EllipsisType
from numpy import ndarray
import astropy.units as U
from martini.datacube import DataCube as DataCube
from martini.sources.sph_source import SPHSource as SPHSource

def find_fwhm(f: T.Callable[[ndarray], float]) -> float: ...

class _BaseSPHKernel(metaclass=abc.ABCMeta):
    __metaclass__: Incomplete

    def __init__(self) -> None: ...
    def _px_weight(
        self, dij: U.Quantity[U.pix], mask: T.Optional[ndarray] = ...
    ) -> U.Quantity[U.pix**2]: ...
    def _confirm_validation(self, noraise: bool = ..., quiet: bool = ...) -> bool: ...
    def _validate_error(
        self,
        msg: str,
        sm_lengths: U.Quantity[U.pix],
        valid: ndarray,
        noraise: bool = ...,
        quiet: bool = ...,
    ) -> None: ...
    def eval_kernel(
        self, r: ndarray | U.Quantity[U.arcsec], h: ndarray | U.Quantity[U.arcsec]
    ) -> U.Quantity | ndarray | float: ...

    sm_lengths: U.Quantity[U.arcsec]
    sm_ranges: ndarray

    def _apply_mask(self, mask: ndarray) -> None: ...
    def _init_sm_lengths(
        self, source: T.Optional[SPHSource] = ..., datacube: T.Optional[DataCube] = ...
    ) -> None: ...
    @abstractmethod
    def kernel(self, q) -> ndarray: ...
    @abstractmethod
    def _kernel_integral(
        self,
        dij: U.Quantity[U.pix],
        h: U.Quantity[U.pix],
        mask: ndarray | EllipsisType | slice = ...,
    ) -> ndarray: ...
    @abstractmethod
    def _validate(
        self, sm_lengths: U.Quantity[U.pix], noraise: bool = ..., quiet: bool = ...
    ) -> ndarray: ...

class _WendlandC2Kernel(_BaseSPHKernel):
    min_valid_size: float
    size_in_fwhm: float

    def __init__(self) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...
    def _kernel_integral(
        self,
        dij: U.Quantity[U.pix],
        h: U.Quantity[U.pix],
        mask: ndarray | EllipsisType | slice = ...,
    ) -> ndarray: ...
    def _validate(
        self, sm_lengths: U.Quantity[U.pix], noraise: bool = ..., quiet: bool = ...
    ) -> ndarray: ...

class _WendlandC6Kernel(_BaseSPHKernel):
    min_valid_size: float
    size_in_fwhm: float

    def __init__(self) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...
    def _kernel_integral(
        self,
        dij: U.Quantity[U.pix],
        h: U.Quantity[U.pix],
        mask: ndarray | EllipsisType | slice = ...,
    ) -> ndarray: ...
    def _validate(
        self, sm_lengths: U.Quantity[U.pix], noraise: bool = ..., quiet: bool = ...
    ) -> ndarray: ...

class _CubicSplineKernel(_BaseSPHKernel):
    min_valid_size: float
    size_in_fwhm: float

    def __init__(self) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...
    def _kernel_integral(
        self,
        dij: U.Quantity[U.pix],
        h: U.Quantity[U.pix],
        mask: ndarray | EllipsisType | slice = ...,
    ) -> ndarray: ...
    def _validate(
        self, sm_lengths: U.Quantity[U.pix], noraise: bool = ..., quiet: bool = ...
    ) -> ndarray: ...

class _GaussianKernel(_BaseSPHKernel):
    no6sigwarn: bool
    truncate: float
    lims: T.Tuple[int, int]
    norm: float
    size_in_fwhm: float

    def __init__(self, truncate: float = ...) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...
    def _kernel_integral(
        self,
        dij: U.Quantity[U.pix],
        h: U.Quantity[U.pix],
        mask: ndarray | EllipsisType | slice = ...,
    ) -> ndarray: ...
    def _validate(
        self, sm_lengths: U.Quantity[U.pix], noraise: bool = ..., quiet: bool = ...
    ) -> ndarray: ...

class DiracDeltaKernel(_BaseSPHKernel):
    max_valid_size: float
    size_in_fwhm: float

    def __init__(self, size_in_fwhm: float = ...) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...
    def _kernel_integral(
        self,
        dij: U.Quantity[U.pix],
        h: U.Quantity[U.pix],
        mask: ndarray | EllipsisType | slice = ...,
    ) -> ndarray: ...
    def _validate(
        self, sm_lengths: U.Quantity[U.pix], noraise: bool = ..., quiet: bool = ...
    ) -> ndarray: ...

class _AdaptiveKernel(_BaseSPHKernel):
    kernels: T.Tuple[_BaseSPHKernel]
    size_in_fwhm: T.Optional[T.Tuple[float]]

    def __init__(self, kernels: T.Tuple[_BaseSPHKernel]) -> None: ...

    kernel_indices: ndarray

    def _init_sm_lengths(
        self, source: T.Optional[SPHSource] = ..., datacube: T.Optional[DataCube] = ...
    ) -> None: ...
    def _apply_mask(self, mask: ndarray) -> None: ...
    def eval_kernel(
        self, r: ndarray | U.Quantity[U.arcsec], h: ndarray | U.Quantity[U.arcsec]
    ) -> U.Quantity | ndarray | float: ...
    def kernel(self, q: ndarray) -> ndarray: ...
    def _kernel_integral(
        self,
        dij: U.Quantity[U.pix],
        h: U.Quantity[U.pix],
        mask: ndarray | EllipsisType | slice = ...,
    ) -> ndarray: ...
    def _validate(
        self, sm_lengths: U.Quantity[U.pix], noraise: bool = ..., quiet: bool = ...
    ) -> ndarray: ...

class _QuarticSplineKernel(_BaseSPHKernel):
    min_valid_size: float
    size_in_fwhm: float

    def __init__(self) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...
    def _kernel_integral(
        self,
        dij: U.Quantity[U.pix],
        h: U.Quantity[U.pix],
        mask: ndarray | EllipsisType | slice = ...,
    ) -> ndarray: ...
    def _validate(
        self, sm_lengths: U.Quantity[U.pix], noraise: bool = ..., quiet: bool = ...
    ) -> ndarray: ...

class WendlandC2Kernel(_AdaptiveKernel):
    def __init__(self) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...

class WendlandC6Kernel(_AdaptiveKernel):
    def __init__(self) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...

class CubicSplineKernel(_AdaptiveKernel):
    def __init__(self) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...

class GaussianKernel(_AdaptiveKernel):
    def __init__(self, truncate: float = ...) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...

class QuarticSplineKernel(_AdaptiveKernel):
    def __init__(self) -> None: ...
    def kernel(self, q: ndarray) -> ndarray: ...

class AdaptiveKernel(object):
    def __init__(self, *args, **kwargs) -> None: ...



from martini._demo import demo, demo_source
import os


class TestDemo:
    def test_demo(self):
        """
        Check that demo completes and produces output files.
        """
        filenames = dict(
            cubefile="pytest_testcube.fits",
            beamfile="pytest_testbeam.fits",
            hdf5file="pytest_testcube.hdf5",
        )
        try:
            demo(**filenames)
        finally:
            for filename in filenames.values():
                if os.path.exists(filename):
                    os.remove(filename)

    def test_demo_source(self):
        """
        Check that demo source created and contains particles.
        """
        N = 100
        source = demo_source(N=N)
        assert source.mHI_g.size == N



"""
Functions which can be used to extend the
:class:`~astropy.coordinates.representation.cartesian.CartesianRepresentation` and
:class:`~astropy.coordinates.representation.cartesian.CartesianDifferential` classes from
:mod:`astropy.coordinates`.

These should be imported, then use:

.. code-block:: python

    # Extend CartesianRepresentation to allow coordinate translation
    setattr(CartesianRepresentation, 'translate', translate)

    # Extend CartesianDifferential to allow velocity (or other differential)
    # translation
    setattr(CartesianDifferential, 'translate', translate_d)

Copied from github.com/kyleaoman/kyleaoman_utilities/kyleaoman_utilities/
commit-id 81e08768bcf3f910d86757c07b44632f393f29aa
Note: No git-based solution (e.g. via submodules) seems practical to include
selected files from external repositories; a direct copy is included here
to produce a self-contained package.
"""

from astropy.coordinates import CartesianRepresentation, CartesianDifferential


def translate(cls, translation_vector):
    """
    Apply a coordinate translation.

    Parameters
    ----------
    translation_vector : ~astropy.units.Quantity
        :class:`~astropy.units.Quantity`, with dimensions of length.
        3-vector by which to translate.

    Returns
    -------
    out : ~astropy.coordinates.representation.cartesian.CartesianRepresentation
        A new
        :class:`~astropy.coordinates.representation.cartesian.CartesianRepresentation`
        instance with translation applied.
    """

    return CartesianRepresentation(
        cls.__class__.get_xyz(cls) + translation_vector.reshape(3, 1),
        differentials=cls.differentials,
    )


def translate_d(cls, translation_vector):
    """
    Apply a differential translation.

    Parameters
    ----------
    translation_vector : ~astropy.units.Quantity
        :class:`~astropy.units.Quantity` with dimensions of velocity (or other
        differential) 3-vector by which to translate.

    Returns
    -------
    out : ~astropy.coordinates.representation.cartesian.CartesianDifferential
        A new
        :class:~astropy.coordinates.representation.cartesian.CartesianDifferential`
        instance with translation applied.
    """

    return CartesianDifferential(
        cls.__class__.get_d_xyz(cls) + translation_vector.reshape(3, 1)
    )



import os
import pytest
import numpy as np
from martini.martini import Martini, GlobalProfile, _BaseMartini
from martini.datacube import DataCube, HIfreq
from martini.beams import GaussianBeam
from test_sph_kernels import simple_kernels
from martini.sph_kernels import _CubicSplineKernel, _GaussianKernel, DiracDeltaKernel
from martini.spectral_models import DiracDeltaSpectrum, GaussianSpectrum
from astropy import units as U
from astropy.io import fits
from astropy import wcs
from astropy.coordinates import FK5, ICRS
from scipy.signal import fftconvolve


def check_mass_accuracy(m, out_mode):
    if out_mode == "hdf5":
        h5py = pytest.importorskip(
            "h5py", reason="h5py (optional dependency) not available"
        )

    # flux in channels
    F = (m.datacube._array * m.datacube.px_size**2).sum((0, 1)).squeeze()  # Jy

    # distance
    D = m.source.distance

    # channel width
    dv = np.abs(np.diff(m.datacube.velocity_channel_edges))

    # HI mass
    MHI = np.sum(
        2.36e5
        * U.Msun
        * D.to_value(U.Mpc) ** 2
        * F.to_value(U.Jy)
        * dv.to_value(U.km / U.s)
    ).to(U.Msun)

    # demand accuracy within 1% after source insertion
    assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

    m.convolve_beam()

    # radiant intensity
    Irad = m.datacube._array.sum((0, 1)).squeeze()  # Jy / beam

    # beam area, for an equivalent Gaussian beam
    A = np.pi * m.beam.bmaj * m.beam.bmin / 4 / np.log(2) / U.beam

    # distance
    D = m.source.distance

    # channel width
    dv = np.abs(np.diff(m.datacube.velocity_channel_edges))

    # flux
    F = (Irad / A).to(U.Jy / U.arcsec**2) * m.datacube.px_size**2

    # HI mass
    MHI = np.sum(
        2.36e5
        * U.Msun
        * D.to_value(U.Mpc) ** 2
        * F.to_value(U.Jy)
        * dv.to_value(U.km / U.s)
    ).to(U.Msun)

    # demand accuracy within 1% after beam convolution
    assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

    if out_mode == "fits":
        filename = "cube.fits"
        try:
            m.write_fits(filename)
            with fits.open(filename) as f:
                # distance
                D = m.source.distance

                # radiant intensity
                fits_wcs = wcs.WCS(f[0].header)
                Irad = U.Quantity(
                    f[0].data.T.sum((0, 1)).squeeze(), unit=f[0].header["BUNIT"]
                )

                A = (
                    np.pi
                    * (f[0].header["BMAJ"] * U.deg)
                    * (f[0].header["BMIN"] * U.deg)
                    / 4
                    / np.log(2)
                    / U.beam
                )
                px_area = U.Quantity(
                    np.abs(f[0].header["CDELT1"]), unit=f[0].header["CUNIT1"]
                ) * U.Quantity(
                    np.abs(f[0].header["CDELT2"]), unit=f[0].header["CUNIT2"]
                )

                # flux
                F = (Irad / A).to(U.Jy / U.arcsec**2) * px_area

                channel_edges = fits_wcs.sub(("spectral",)).all_pix2world(
                    np.arange(fits_wcs.sub(("spectral",)).pixel_shape[0] + 1) - 0.5,
                    0,
                ) * U.Unit(fits_wcs.wcs.cunit[fits_wcs.wcs.spec], format="fits")
                dv = np.abs(
                    np.diff(
                        channel_edges.squeeze().to(
                            U.km / U.s, equivalencies=U.doppler_radio(HIfreq)
                        )
                    )
                )

                # HI mass
                MHI = np.sum(
                    2.36e5
                    * U.Msun
                    * D.to_value(U.Mpc) ** 2
                    * F.to_value(U.Jy)
                    * dv.to_value(U.km / U.s)
                ).to(U.Msun)

            # demand accuracy within 1% in output fits file
            assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

        finally:
            if os.path.exists(filename):
                os.remove(filename)

    if out_mode == "hdf5":
        filename = "cube.hdf5"
        try:
            m.write_hdf5(filename)
            with h5py.File(filename, "r") as f:
                # distance
                D = m.source.distance

                # radiant intensity
                Irad = U.Quantity(
                    f["FluxCube"][()].sum((0, 1)).squeeze(),
                    unit=f["FluxCube"].attrs["FluxCubeUnit"],
                )

                A = (
                    np.pi
                    * (f["FluxCube"].attrs["BeamMajor_in_deg"] * U.deg)
                    * (f["FluxCube"].attrs["BeamMinor_in_deg"] * U.deg)
                    / 4
                    / np.log(2)
                    / U.beam
                )
                dv = np.abs(
                    np.diff(
                        f["velocity_channel_edges"]
                        * U.Unit(f["velocity_channel_edges"].attrs["Unit"])
                    )
                )
                px_area = U.Quantity(
                    np.abs(f["FluxCube"].attrs["deltaRA_in_RAUnit"]),
                    unit=f["FluxCube"].attrs["RAUnit"],
                ) * U.Quantity(
                    np.abs(f["FluxCube"].attrs["deltaDec_in_DecUnit"]),
                    unit=f["FluxCube"].attrs["DecUnit"],
                )

                # flux
                F = (Irad / A).to(U.Jy / U.arcsec**2) * px_area

                # HI mass
                MHI = np.sum(
                    2.36e5
                    * U.Msun
                    * D.to_value(U.Mpc) ** 2
                    * F.to_value(U.Jy)
                    * dv.to_value(U.km / U.s)
                ).to(U.Msun)

            # demand accuracy within 1% in output hdf5 file
            assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

        finally:
            if os.path.exists(filename):
                os.remove(filename)


class TestMartini:
    @pytest.mark.parametrize("sph_kernel", simple_kernels)
    @pytest.mark.parametrize("spectral_model", (DiracDeltaSpectrum, GaussianSpectrum))
    @pytest.mark.parametrize("out_mode", ("fits", "hdf5"))
    def test_mass_accuracy(
        self, dc_zeros, sph_kernel, spectral_model, single_particle_source, out_mode
    ):
        """
        Check that the input mass in the particles gives the correct total mass in the
        datacube, by checking the conversion back to total mass. Covers testing
        Martini.insert_source_in_cube.
        """

        hsm_g = (
            0.1 * U.kpc if sph_kernel.__name__ == "DiracDeltaKernel" else 1.0 * U.kpc
        )
        source = single_particle_source(hsm_g=hsm_g)
        # single_particle_source has a mass of 1E4Msun, temperature of 1E4K
        m = Martini(
            source=source,
            datacube=dc_zeros,
            beam=GaussianBeam(),
            noise=None,
            spectral_model=spectral_model(),
            sph_kernel=sph_kernel(),
        )
        m.insert_source_in_cube(progressbar=False)
        check_mass_accuracy(m, out_mode)

    def test_convolve_beam(self, single_particle_source):
        """
        Check that beam convolution gives result matching manual calculation.
        """
        source = single_particle_source()
        datacube = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=16,
            spectral_centre=source.distance * source.h * 100 * U.km / U.s / U.Mpc,
        )
        beam = GaussianBeam()
        noise = None
        sph_kernel = _GaussianKernel()
        spectral_model = GaussianSpectrum()

        m = Martini(
            source=source,
            datacube=datacube,
            beam=beam,
            noise=noise,
            sph_kernel=sph_kernel,
            spectral_model=spectral_model,
        )
        m.insert_source_in_cube()
        unconvolved_cube = m.datacube._array.copy()
        unit = unconvolved_cube.unit
        s = np.s_[..., 0] if m.datacube.stokes_axis else np.s_[...]
        for spatial_slice in iter(unconvolved_cube[s].transpose((2, 0, 1))):
            spatial_slice[...] = (
                fftconvolve(spatial_slice, m.beam.kernel, mode="same") * unit
            )
        convolved_cube = unconvolved_cube[
            m.datacube.padx : -m.datacube.padx, m.datacube.pady : -m.datacube.padx
        ]
        convolved_cube = convolved_cube.to(
            U.Jy * U.beam**-1,
            equivalencies=U.beam_angular_area(m.beam.area),
        )
        m.convolve_beam()
        assert U.allclose(m.datacube._array, convolved_cube)

    def test_add_noise(self, m_init):
        """
        Check that noise provided goes into the datacube when we call add_noise.
        """
        assert (m_init.datacube._array.sum() == 0).all()
        assert m_init.noise.seed is not None
        expected_noise = m_init.noise.generate(m_init.datacube, m_init.beam)
        m_init.noise.reset_rng()
        m_init.add_noise()
        assert U.allclose(
            m_init.datacube._array,
            expected_noise.to(
                U.Jy * U.arcsec**-2,
                equivalencies=U.beam_angular_area(m_init.beam.area),
            ).to(
                m_init.datacube._array.unit,
                equivalencies=[m_init.datacube.arcsec2_to_pix],
            ),
        )

    @pytest.mark.parametrize(
        ("ra_off", "ra_in"),
        (
            (0 * U.arcsec, True),
            (3 * U.arcsec, True),
            (9 * U.arcsec, False),
            (-3 * U.arcsec, True),
            (-9 * U.arcsec, False),
        ),
    )
    @pytest.mark.parametrize(
        ("dec_off", "dec_in"),
        (
            (0 * U.arcsec, True),
            (3 * U.arcsec, True),
            (9 * U.arcsec, False),
            (-3 * U.arcsec, True),
            (-9 * U.arcsec, False),
        ),
    )
    @pytest.mark.parametrize(
        ("v_off", "v_in"),
        (
            (0 * U.km / U.s, True),
            (3 * U.km / U.s, True),
            (7 * U.km / U.s, False),
            (-3 * U.km / U.s, True),
            (-7 * U.km / U.s, False),
        ),
    )
    @pytest.mark.parametrize("spatial", (True, False))
    @pytest.mark.parametrize("spectral", (True, False))
    def test_prune_particles(
        self,
        ra_off,
        ra_in,
        dec_off,
        dec_in,
        v_off,
        v_in,
        single_particle_source,
        spatial,
        spectral,
    ):
        """
        Check that a particle offset by a specific set of (RA, Dec, v) is inside/outside
        the cube as expected.
        """
        if spatial and spectral:
            expect_particle = all((ra_in, dec_in, v_in))
        elif spatial and not spectral:
            expect_particle = all((ra_in, dec_in))
        elif spectral and not spatial:
            expect_particle = v_in
        elif not spectral and not spatial:
            expect_particle = True
        # set distance so that 1kpc = 1arcsec
        distance = (1 * U.kpc / 1 / U.arcsec).to(U.Mpc, U.dimensionless_angles())
        source = single_particle_source(
            distance=distance, ra=ra_off, dec=dec_off, vpeculiar=v_off
        )
        datacube = DataCube(
            n_px_x=2,
            n_px_y=2,
            n_channels=2,
            spectral_centre=source.distance * source.h * 100 * U.km / U.s / U.Mpc,
            px_size=1 * U.arcsec,
            channel_width=1 * U.km / U.s,
            ra=0 * U.deg,
            dec=0 * U.deg,
        )
        # pad size will be 5, so datacube is 12x12 pixels
        beam = GaussianBeam(bmaj=1 * U.arcsec, bmin=1 * U.arcsec, truncate=4)
        sph_kernel = _CubicSplineKernel()
        spectral_model = GaussianSpectrum(sigma=1 * U.km / U.s)
        # need to use _BaseMartini below to manipulate _prune_kwargs
        kwargs = dict(
            source=source,
            datacube=datacube,
            beam=beam,
            noise=None,
            sph_kernel=sph_kernel,
            spectral_model=spectral_model,
            _prune_kwargs=dict(spatial=spatial, spectral=spectral),
        )
        # if more than 1px (datacube) + 5px (pad) + 2px (sm_range) then expect to prune
        # if more than 1px (datacube) + 4px (4*spectrum_half_width) then expect to prune
        if not expect_particle:
            with pytest.raises(
                RuntimeError, match="No source particles in target region."
            ):
                _BaseMartini(**kwargs)
        else:
            assert _BaseMartini(**kwargs).source.npart == 1

    def test_reset(self, m_nn):
        """
        Check that resetting martini instance zeros out datacube.
        """
        cube_array = m_nn.datacube._array
        assert m_nn.datacube._array.sum() > 0
        m_nn.reset()
        assert m_nn.datacube._array.sum() == 0
        # check that can start over and get the same result w/o errors
        m_nn.insert_source_in_cube(progressbar=False)
        m_nn.convolve_beam()
        assert U.allclose(cube_array, m_nn.datacube._array)
        # check that can reset after doing nothing
        m_nn.reset()
        m_nn.reset()

    def test_reset_preserves_shape(self, single_particle_source, dc_zeros):
        m = Martini(
            source=single_particle_source(),
            datacube=dc_zeros,
            beam=GaussianBeam(),
            noise=None,
            spectral_model=DiracDeltaSpectrum(),
            sph_kernel=DiracDeltaKernel(),
        )
        expected_shape = m.datacube._array.shape
        m.reset()
        assert m.datacube._array.shape == expected_shape

    def test_preview(self, m_init):
        """
        Simply check that the preview visualisation runs without error.
        """
        pytest.importorskip(
            "matplotlib", reason="matplotlib (optional dependency) not available."
        )
        # with default arguments
        with pytest.warns(UserWarning, match="singular"):
            # warning: single-particle source is used, so axis limits try to be equal
            m_init.preview()
        # with non-default arguments
        m_init.preview(
            max_points=1000,
            fig=2,
            lim="datacube",
            vlim="datacube",
            point_scaling="fixed",
            title="test",
        )

    def test_source_to_datacube_coord_transformation(self, single_particle_source):
        """
        Check that transformation is applied if source and datacube have different
        coordinate frames.
        """
        source = single_particle_source(hsm_g=0.01 * U.kpc)
        assert source.coordinate_frame.name == "icrs"
        datacube_icrs = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=16,
            channel_width=4 * U.km / U.s,
            px_size=10 * U.arcsec,
            spectral_centre=source.vsys,
            ra=source.ra,
            dec=source.dec,
            coordinate_frame=ICRS(),
        )
        m_icrs = Martini(
            source=source,
            datacube=datacube_icrs,
            beam=GaussianBeam(),
            noise=None,
            sph_kernel=DiracDeltaKernel(),
            spectral_model=DiracDeltaSpectrum(),
        )

        def centre_pixels_slice(m):
            datacube = m.datacube
            return m.datacube._array[
                datacube.n_px_x // 2
                - 1
                + datacube.padx : datacube.n_px_x // 2
                + 1
                + datacube.padx,
                datacube.n_px_y // 2
                - 1
                + datacube.pady : datacube.n_px_y // 2
                + 1
                + datacube.pady,
            ]

        assert np.sum(centre_pixels_slice(m_icrs).sum()) == 0
        m_icrs.insert_source_in_cube(progressbar=False)
        assert np.sum(centre_pixels_slice(m_icrs).sum()) > 0

        # ICRS is ~J2000 equinox. J1950 equinox is about a degree off,
        # so we should completely miss the cube (16 pix of 10 arcsec).
        datacube_fk5_J1950 = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=16,
            channel_width=4 * U.km / U.s,
            px_size=10 * U.arcsec,
            spectral_centre=source.vsys,
            ra=source.ra,
            dec=source.dec,
            coordinate_frame=FK5(equinox="J1950"),
        )
        with pytest.raises(RuntimeError, match="No source particles in target region."):
            Martini(
                source=source,
                datacube=datacube_fk5_J1950,
                beam=GaussianBeam(),
                noise=None,
                sph_kernel=DiracDeltaKernel(),
                spectral_model=DiracDeltaSpectrum(),
            )

    def test_source_to_datacube_specsys_transformation(self, single_particle_source):
        """
        Check that spectral reference transformation is applied if source and datacube
        have different specsys.
        """
        source = single_particle_source(hsm_g=0.01 * U.kpc)
        datacube_icrs = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=16,
            channel_width=4 * U.km / U.s,
            px_size=10 * U.arcsec,
            spectral_centre=source.vsys,
            ra=source.ra,
            dec=source.dec,
            coordinate_frame=ICRS(),
            specsys="icrs",
        )
        m_icrs = Martini(
            source=source,
            datacube=datacube_icrs,
            beam=GaussianBeam(),
            noise=None,
            sph_kernel=DiracDeltaKernel(),
            spectral_model=DiracDeltaSpectrum(),
        )

        def centre_channels_slice(m):
            datacube = m.datacube
            return m.datacube._array[
                :, :, datacube.n_channels // 2 - 1 : datacube.n_channels // 2 + 1
            ]

        assert np.sum(centre_channels_slice(m_icrs).sum()) == 0
        m_icrs.insert_source_in_cube(progressbar=False)
        assert np.sum(centre_channels_slice(m_icrs).sum()) > 0

        # ICRS and Galactocentric are offset by many km/s depending on direction
        # so with 4 channels of 1 km/s we should completely miss the cube
        datacube_galactocentric = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=4,
            channel_width=1 * U.km / U.s,
            px_size=10 * U.arcsec,
            spectral_centre=source.vsys,
            ra=source.ra,
            dec=source.dec,
            coordinate_frame=ICRS(),
            specsys="galactocentric",
        )
        assert datacube_galactocentric.wcs.wcs.specsys == "galactocentric"
        with pytest.raises(RuntimeError, match="No source particles in target region."):
            Martini(
                source=source,
                datacube=datacube_galactocentric,
                beam=GaussianBeam(),
                noise=None,
                sph_kernel=DiracDeltaKernel(),
                spectral_model=DiracDeltaSpectrum(),
            )

    def test_hdf5_grids(self, m):
        h5py = pytest.importorskip(
            "h5py", reason="h5py (optional dependency) not available."
        )
        origin = 0  # numpy-like, not fits-like, indexing
        filename = "cube.hdf5"
        try:
            m.write_hdf5(filename)
            with h5py.File(filename, "r") as f:
                voxel_coords = np.vstack(
                    (
                        (
                            f["RA"][()].flatten()
                            * U.Unit(f["RA"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[0]),
                        (
                            f["Dec"][()].flatten()
                            * U.Unit(f["Dec"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[1]),
                        (
                            f["channel_mids"][()].flatten()
                            * U.Unit(f["channel_mids"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[2]),
                    )
                ).T
                ra_idx, dec_idx, spec_idx = m.datacube.wcs.all_world2pix(
                    voxel_coords, origin
                ).T
                ra_idx = ra_idx.reshape((m.datacube._array.shape))
                dec_idx = dec_idx.reshape((m.datacube._array.shape))
                spec_idx = spec_idx.reshape((m.datacube._array.shape))
                expected_idx = np.meshgrid(
                    np.arange(m.datacube.n_px_x),
                    np.arange(m.datacube.n_px_y),
                    np.arange(m.datacube.n_channels),
                    indexing="ij",
                )
                assert np.allclose(ra_idx, expected_idx[0])
                assert np.allclose(dec_idx, expected_idx[1])
                assert np.allclose(spec_idx, expected_idx[2])
                vertex_coords = np.vstack(
                    (
                        (
                            f["RA_vertices"][()].flatten()
                            * U.Unit(f["RA_vertices"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[0]),
                        (
                            f["Dec_vertices"][()].flatten()
                            * U.Unit(f["Dec_vertices"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[1]),
                        (
                            f["channel_vertices"][()].flatten()
                            * U.Unit(f["channel_vertices"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[2]),
                    )
                ).T
                ra_vx_idx, dec_vx_idx, spec_vx_idx = m.datacube.wcs.all_world2pix(
                    vertex_coords, origin
                ).T
                shape = [s + 1 for s in m.datacube._array.shape]
                ra_vx_idx = ra_vx_idx.reshape(shape)
                dec_vx_idx = dec_vx_idx.reshape(shape)
                spec_vx_idx = spec_vx_idx.reshape(shape)
                expected_vx_idx = np.meshgrid(
                    np.arange(m.datacube.n_px_x + 1) - 0.5,
                    np.arange(m.datacube.n_px_y + 1) - 0.5,
                    np.arange(m.datacube.n_channels + 1) - 0.5,
                    indexing="ij",
                )
                assert np.allclose(ra_vx_idx, expected_vx_idx[0])
                assert np.allclose(dec_vx_idx, expected_vx_idx[1])
                assert np.allclose(spec_vx_idx, expected_vx_idx[2])

        finally:
            if os.path.exists(filename):
                os.remove(filename)


class TestParallel:
    def test_parallel_consistent_with_serial(self, many_particle_source, dc_zeros):
        """
        Check that running the source insertion loop in parallel gives the same result
        as running in serial.
        """
        pytest.importorskip(
            "multiprocess", reason="multiprocess (optional dependency) not available"
        )

        m = Martini(
            source=many_particle_source(),
            datacube=dc_zeros,
            beam=GaussianBeam(),
            noise=None,
            sph_kernel=_GaussianKernel(),
            spectral_model=GaussianSpectrum(),
        )

        m.insert_source_in_cube(ncpu=1, progressbar=False)
        expected_result = m.datacube._array

        # check that we're not testing on a zero array
        assert m.datacube._array.sum() > 0

        m.reset()

        # check the reset was successful
        assert np.allclose(
            m.datacube._array.to_value(m.datacube._array.unit),
            0.0,
        )

        m.insert_source_in_cube(ncpu=2, progressbar=False)

        assert U.allclose(m.datacube._array, expected_result)


class TestGlobalProfile:
    @pytest.mark.parametrize("spectral_model", (DiracDeltaSpectrum, GaussianSpectrum))
    @pytest.mark.parametrize("ra", (0 * U.deg, 180 * U.deg))
    def test_mass_accuracy(self, spectral_model, single_particle_source, ra):
        """
        Check that the input mass in the particles gives the correct total mass in the
        spectrum, by checking the conversion back to total mass. Covers testing
        GlobalProfile.insert_source_in_spectrum.
        """

        # single_particle_source has a mass of 1E4Msun, temperature of 1E4K
        # we test both ra=0deg and ra=180deg to make sure all particles included
        source = single_particle_source(ra=ra)
        m = GlobalProfile(
            source=source,
            spectral_model=spectral_model(),
            n_channels=32,
            channel_width=10 * U.km * U.s**-1,
            spectral_centre=source.vsys,
        )

        m.insert_source_in_spectrum()

        # flux
        F = m.spectrum.sum()  # Jy

        # distance
        D = m.source.distance

        # channel width
        dv = m.channel_width

        # HI mass
        MHI = (
            2.36e5
            * U.Msun
            * D.to_value(U.Mpc) ** 2
            * F.to_value(U.Jy)
            * dv.to_value(U.km / U.s)
        ).to(U.Msun)

        # demand accuracy within 1% after source insertion
        assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

    @pytest.mark.parametrize(
        ("ra_off", "ra_in"),
        (
            (0 * U.arcsec, True),
            (3 * U.arcsec, True),
            (5 * U.deg, False),  # global profile uses 1 deg pixel
            (-3 * U.arcsec, True),
            (-5 * U.deg, False),  # global profile uses 1 deg pixel
        ),
    )
    @pytest.mark.parametrize(
        ("dec_off", "dec_in"),
        (
            (0 * U.arcsec, True),
            (3 * U.arcsec, True),
            (5 * U.deg, False),  # global profile uses 1 deg pixel
            (-3 * U.arcsec, True),
            (-5 * U.deg, False),  # global profile uses 1 deg pixel
        ),
    )
    @pytest.mark.parametrize(
        ("v_off", "v_in"),
        (
            (0 * U.km / U.s, True),
            (3 * U.km / U.s, True),
            (7 * U.km / U.s, False),
            (-3 * U.km / U.s, True),
            (-7 * U.km / U.s, False),
        ),
    )
    def test_prune_particles(
        self, ra_off, ra_in, dec_off, dec_in, v_off, v_in, single_particle_source
    ):
        """
        Check that a particle offset by a specific set of (RA, Dec, v) is inside/outside
        the cube as expected. GlobalProfile should ignore RA, Dec when pruning.
        """
        # GlobalProfile should ignore RA, Dec when pruning:
        expect_particle = v_in
        # set distance so that 1kpc = 1arcsec
        distance = (1 * U.kpc / 1 / U.arcsec).to(U.Mpc, U.dimensionless_angles())
        source = single_particle_source(
            distance=distance, ra=ra_off, dec=dec_off, vpeculiar=v_off
        )
        spectral_model = GaussianSpectrum(sigma=1 * U.km / U.s)
        kwargs = dict(
            source=source,
            spectral_model=spectral_model,
            n_channels=2,
            channel_width=1 * U.km / U.s,
            spectral_centre=source.distance * source.h * 100 * U.km / U.s / U.Mpc,
        )
        # if more than 1px (datacube) + 4px (4*spectrum_half_width) then expect to prune
        if not expect_particle:
            with pytest.raises(
                RuntimeError, match="No source particles in target region."
            ):
                GlobalProfile(**kwargs)
        else:
            assert GlobalProfile(**kwargs).source.npart == 1

    def test_reset(self, gp):
        """
        Check that resetting global profile instance zeros out datacube and spectrum.
        """
        cube_array = gp._datacube._array
        assert gp._datacube._array.sum() > 0
        spectrum = gp.spectrum
        assert spectrum.sum() > 0
        gp.reset()
        assert gp._datacube._array.sum() == 0
        assert not hasattr(gp, "_spectrum")
        # check that can start over and get the same result w/o errors
        gp.insert_source_in_spectrum()
        assert U.allclose(cube_array, gp._datacube._array)
        assert U.allclose(spectrum, gp.spectrum)
        # check that can reset after doing nothing
        gp.reset()
        gp.reset()

    def test_preview(self, gp):
        """
        Simply check that the preview visualisation runs without error.
        """
        pytest.importorskip(
            "matplotlib", reason="matplotlib (optional dependency) not available."
        )
        # with default arguments
        with pytest.warns(UserWarning, match="singular"):
            # warning: single-particle source is used, so axis limits try to be equal
            gp.preview()
        # with non-default arguments
        gp.preview(
            max_points=1000,
            fig=2,
            lim="datacube",
            vlim="datacube",
            point_scaling="fixed",
            title="test",
        )

    def test_channel_modes(self, single_particle_source):
        """
        Check that channels have expected units in both modes (frequency, velocity).
        """
        source = single_particle_source()
        channel_width = 10 * U.km * U.s**-1
        m = GlobalProfile(
            source=source,
            spectral_model=GaussianSpectrum(sigma="thermal"),
            n_channels=32,
            channel_width=channel_width,
            spectral_centre=source.vsys,
        )
        expected_units = channel_width.unit
        # these will raise if there's a problem:
        m.channel_edges.to(expected_units)
        m.channel_mids.to(expected_units)

    def test_view_spectrum(self, gp):
        """
        Simply check that plotting spectrum runs without error.
        """
        pytest.importorskip(
            "matplotlib", reason="matplotlib (optional dependency) not available."
        )
        # with default arguments
        gp.plot_spectrum()
        # with non-default arguments
        gp.plot_spectrum(fig=2, title="test", show_vsys=False)


class TestMartiniWithDataCubeFromWCS:

    @pytest.mark.parametrize("out_mode", ("fits", "hdf5"))
    def test_source_insertion(self, dc_wcs, single_particle_source, out_mode):
        datacube = dc_wcs
        distance = (
            datacube.spectral_centre.to(
                U.km / U.s, equivalencies=U.doppler_radio(HIfreq)
            )
            / (70 * U.km / U.s / U.Mpc)
        ).to(U.Mpc)
        source = single_particle_source(
            ra=datacube.ra,
            dec=datacube.dec,
            distance=distance,
            hsm_g=(3 * datacube.px_size * distance).to(
                U.kpc, equivalencies=U.dimensionless_angles()
            ),
        )
        beam = GaussianBeam(bmaj=3 * datacube.px_size, bmin=3 * datacube.px_size)
        m = Martini(
            source=source,
            datacube=datacube,
            beam=beam,
            noise=None,
            spectral_model=GaussianSpectrum(sigma="thermal"),
            sph_kernel=_CubicSplineKernel(),
        )
        m.insert_source_in_cube(progressbar=False)
        check_mass_accuracy(m, out_mode)



from numpy import ndarray
import typing as T
from martini.beams import _BaseBeam
from martini.datacube import DataCube as DataCube
from martini.noise import _BaseNoise
from martini.sources.sph_source import SPHSource as SPHSource
from martini.spectral_models import _BaseSpectrum
from martini.sph_kernels import _BaseSPHKernel
from matplotlib.figure import Figure
import astropy.units as U

gc: bytes
martini_version: str

class _BaseMartini:
    source: SPHSource
    _datacube: DataCube
    beam: _BaseBeam
    noise: _BaseNoise
    sph_kernel: _BaseSPHKernel
    spectral_model: _BaseSpectrum
    quiet: bool

    def __init__(
        self,
        source: T.Optional[SPHSource] = ...,
        datacube: T.Optional[DataCube] = ...,
        beam: T.Optional[_BaseBeam] = ...,
        noise: T.Optional[_BaseNoise] = ...,
        sph_kernel: T.Optional[_BaseSPHKernel] = ...,
        spectral_model: T.Optional[_BaseSpectrum] = ...,
        quiet: T.Optional[bool] = ...,
        _prune_kwargs: T.Dict[str, T.Union[bool, str]] = ...,
    ) -> None: ...
    def _prune_particles(
        self, spatial: bool = ..., spectral: bool = ..., obj_type_str: str = ...
    ) -> None: ...
    def _evaluate_pixel_spectrum(
        self,
        ranks_and_ij_pxs: T.Tuple[int, T.List[T.Tuple[int, int]]],
        progressbar: bool = ...,
    ) -> T.Tuple[slice, U.Quantity[U.Jy / U.arcsec**2]]: ...
    def _insert_pixel(
        self, insertion_slice: T.Union[int, T.Tuple, slice], insertion_data: ndarray
    ) -> None: ...
    def _insert_source_in_cube(
        self,
        skip_validation: bool = ...,
        progressbar: T.Optional[bool] = ...,
        ncpu: int = ...,
        quiet: T.Optional[bool] = ...,
    ) -> None: ...
    def reset(self) -> None: ...
    def preview(
        self,
        max_points: int = ...,
        fig: int = ...,
        lim: T.Optional[T.Union[str, U.Quantity[U.kpc]]] = ...,
        vlim: T.Optional[T.Union[str, U.Quantity[U.km / U.s]]] = ...,
        point_scaling: str = ...,
        title: str = ...,
        save: T.Optional[str] = ...,
    ) -> Figure: ...

class Martini:
    def __init__(
        self,
        source: T.Optional[SPHSource] = ...,
        datacube: T.Optional[DataCube] = ...,
        beam: T.Optional[_BaseBeam] = ...,
        noise: T.Optional[_BaseNoise] = ...,
        sph_kernel: T.Optional[_BaseSPHKernel] = ...,
        spectral_model: T.Optional[_BaseSpectrum] = ...,
        quiet: T.Optional[bool] = ...,
    ) -> None: ...
    @property
    def datacube(self) -> DataCube: ...
    def insert_source_in_cube(
        self,
        skip_validation: bool = ...,
        progressbar: T.Optional[bool] = ...,
        ncpu: int = ...,
    ) -> None: ...
    def convolve_beam(self) -> None: ...
    def add_noise(self) -> None: ...
    def write_fits(
        self,
        filename: str,
        overwrite: bool = ...,
        obj_name: str = ...,
        channels: None = ...,
    ) -> None: ...
    def write_beam_fits(
        self, filename: str, overwrite: bool = ..., channels: None = ...
    ) -> None: ...
    def write_hdf5(
        self,
        filename: str,
        overwrite: bool = ...,
        memmap: bool = ...,
        compact: bool = ...,
        channels: None = ...,
    ) -> None: ...

class GlobalProfile(_BaseMartini):
    def __init__(
        self,
        source: T.Optional[SPHSource] = ...,
        spectral_model: T.Optional[_BaseSpectrum] = ...,
        n_channels: int = ...,
        channel_width: U.Quantity[U.km / U.s] = ...,
        spectral_centre: U.Quantity[U.km / U.s] = ...,
        quiet: bool = ...,
        channels: None = ...,
    ) -> None: ...
    def insert_source_in_spectrum(self) -> None: ...
    @property
    def spectrum(self) -> U.Quantity[U.Jy]: ...
    @property
    def channel_edges(self) -> T.Union[U.Quantity[U.Hz], U.Quantity[U.km / U.s]]: ...
    @property
    def channel_mids(self) -> T.Union[U.Quantity[U.Hz], U.Quantity[U.km / U.s]]: ...
    @property
    def velocity_channel_edges(self) -> U.Quantity[U.km / U.s]: ...
    @property
    def velocity_channel_mids(self) -> U.Quantity[U.km / U.s]: ...
    @property
    def frequency_channel_edges(self) -> U.Quantity[U.Hz]: ...
    @property
    def frequency_channel_mids(self) -> U.Quantity[U.Hz]: ...
    @property
    def channel_width(self) -> T.Union[U.Quantity[U.Hz], U.Quantity[U.km / U.s]]: ...
    def reset(self) -> None: ...
    def plot_spectrum(
        self,
        fig: int = ...,
        title: str = ...,
        channels: str = ...,
        show_vsys: bool = ...,
        save: T.Optional[str] = ...,
    ) -> Figure: ...



import os
from martini.__version__ import __version__
import pytest


class TestExamples:
    def test_source_injection_example_version(self):
        """
        Check that the source injection notebook installs current version of martini.
        """
        with open("examples/martini_source_injection.ipynb") as f:
            nb_content = f.read()
        print(nb_content)
        assert (
            '"!{sys.executable} -m pip install \\"astromartini[tngsource]=='
            + __version__
            + '\\"'
            in nb_content
        )

    @pytest.mark.skipif(
        not os.path.isfile("examples/NGC_2841_NA_CUBE_THINGS.FITS"),
        reason="sample data not locally available",
    )
    def test_source_injection_example(self):
        pytest.importorskip(
            "nbmake", reason="nbmake (optional dependency) not available"
        )
        from nbmake.nb_run import NotebookRun
        from nbmake.pytest_items import NotebookFailedException
        import pathlib

        files_to_cleanup = (
            "examples/NGC_2841_AND_TNG50-1_99_737963.FITS",
            "examples/sourceinjectiondemo.fits",
            "examples/martini-cutout-grnr-TNG50-1-99-737963.npy",
            "examples/martini-cutout-TNG50-1-99-2122.hdf5",
        )

        assert os.path.isfile("examples/NGC_2841_NA_CUBE_THINGS.FITS")
        for file_to_cleanup in files_to_cleanup:
            if os.path.isfile(file_to_cleanup):
                os.remove(file_to_cleanup)
        nbr = NotebookRun(pathlib.Path("examples/martini_source_injection.ipynb"), 3600)
        try:
            result = nbr.execute()
            if result.error is not None:
                raise NotebookFailedException(result)
        finally:
            for file_to_cleanup in files_to_cleanup:
                if os.path.isfile(file_to_cleanup):
                    os.remove(file_to_cleanup)



import pytest
import numpy as np
from martini import Martini, DataCube
from martini.spectral_models import GaussianSpectrum
from martini.sources import SPHSource
from martini.sph_kernels import (
    _WendlandC2Kernel,
    _GaussianKernel,
    _CubicSplineKernel,
    _QuarticSplineKernel,
    _WendlandC6Kernel,
    DiracDeltaKernel,
    _AdaptiveKernel,
    WendlandC2Kernel,
    GaussianKernel,
    CubicSplineKernel,
    QuarticSplineKernel,
    WendlandC6Kernel,
)
from astropy import units as U

# kernels that have a well-defined FWHM, i.e. not dirac-delta, adaptive
fwhm_kernels = (
    _WendlandC2Kernel,
    _WendlandC6Kernel,
    _CubicSplineKernel,
    _GaussianKernel,
    _QuarticSplineKernel,
)
simple_kernels = fwhm_kernels + (DiracDeltaKernel,)
recommended_kernels = (
    WendlandC2Kernel,
    WendlandC6Kernel,
    CubicSplineKernel,
    GaussianKernel,
    QuarticSplineKernel,
)
adaptive_kernels = recommended_kernels + (_AdaptiveKernel,)
all_kernels = simple_kernels + adaptive_kernels


def total_kernel_weight(k, h, ngrid=50):
    r = np.arange(0, ngrid)
    dr = (r[1] - r[0]) * U.pix
    xgrid, ygrid = np.meshgrid(np.r_[r[1:][::-1], r], np.r_[r[1:][::-1], r])
    dij = np.vstack((xgrid.flatten(), ygrid.flatten()))
    k.sm_lengths = np.ones(dij.shape[1]) * h * U.pix
    return (dr**2 * np.sum(k._px_weight(dij * U.pix))).to_value(
        U.dimensionless_unscaled
    )


class TestSPHKernels:
    @pytest.mark.parametrize("kernel", fwhm_kernels)
    def test_fwhm_is_one(self, kernel):
        """
        Check that value at FWHM is half of peak value.
        """
        k = kernel()
        fwhm = 1  # all kernels should be implemented s.t. this is true
        assert np.isclose(k.eval_kernel(fwhm / 2, 1), k.eval_kernel(0, 1) / 2)

    @pytest.mark.parametrize("kernel", fwhm_kernels)
    def test_extent(self, kernel):
        """
        Check that kernel goes to zero at its stated size.
        """
        k = kernel()
        fwhm = 1  # all kernels should be implemented s.t. this is true
        assert k.eval_kernel(fwhm * k.size_in_fwhm + 1.0e-5, 1) == 0
        assert k.eval_kernel(fwhm * k.size_in_fwhm - 1.0e-5, 1) > 0

    @pytest.mark.parametrize("kernel", fwhm_kernels)
    def test_2D_integral(self, kernel):
        """
        Check numerically that integral of 3D kernel and 2D projection agree.
        """
        x_2d = list()
        y_2d = list()
        x_3d = list()
        y_3d = list()
        k = kernel()
        vmax = 50
        h = 25
        r = np.arange(0, vmax)
        rmid = 0.5 * (r[1:] + r[:-1])
        dr = r[1] - r[0]
        xgrid, ygrid = np.meshgrid(np.r_[r[1:][::-1], r], np.r_[r[1:][::-1], r])
        rgrid = np.sqrt(np.power(xgrid, 2) + np.power(ygrid, 2))
        for ri in rmid:
            eval_grid = rgrid <= ri
            k.sm_lengths = h * np.ones(rgrid.shape)[eval_grid].flatten() * U.pix
            dij = np.vstack((xgrid[eval_grid], ygrid[eval_grid]))
            IKi = dr**2 * np.sum(
                k._px_weight(
                    dij * U.pix,
                )
            )
            x_2d.append(ri / h)
            y_2d.append(IKi.to_value(U.pix**-2))

        r = r / h
        rmid = 0.5 * (r[1:] + r[:-1])
        dr = r[1] - r[0]
        xgrid, ygrid, zgrid = np.meshgrid(
            np.r_[r[1:][::-1], r], np.r_[r[1:][::-1], r], np.r_[r[1:][::-1], r]
        )
        rgrid = np.sqrt(np.power(xgrid, 2) + np.power(ygrid, 2) + np.power(zgrid, 2))
        Rgrid = np.sqrt(np.power(xgrid, 2) + np.power(ygrid, 2))
        for ri in rmid:
            eval_grid = Rgrid <= ri
            x_3d.append(ri)
            y_3d.append(dr**3 * np.sum(k.eval_kernel(rgrid[eval_grid], 1)))

        assert np.allclose(x_2d, x_3d)  # we evaluated integrals at same radii
        assert np.allclose(y_2d, y_3d, rtol=2.0e-2)  # integrals within 2%

    @pytest.mark.parametrize(
        "kernel",
        (
            _WendlandC2Kernel,
            _WendlandC6Kernel,
            _CubicSplineKernel,
            _QuarticSplineKernel,
        ),
    )
    def test_kernel_validation_minsize(self, kernel):
        """
        Check that our kernel integral approximations hold within the stated
        tolerance, in other words that we conserve mass within the same
        tolerance.
        """
        k = kernel()
        # check that a very well-sampled case gives 1.0
        assert np.isclose(total_kernel_weight(k, 20), 1.0, rtol=1.0e-3)
        # check that the minimum size gives 1.0 within 1%
        assert np.isclose(
            total_kernel_weight(k, kernel.min_valid_size * 1.001), 1.0, rtol=1.0e-2
        )
        assert not np.isclose(
            total_kernel_weight(k, kernel.min_valid_size * 0.9), 1.0, rtol=1.0e-2
        )

    @pytest.mark.parametrize("kernel", (DiracDeltaKernel,))
    def test_kernel_validation_maxsize(self, kernel):
        """
        Check that our kernel integral approximations hold within the stated
        tolerance, in other words that we conserve mass within the same
        tolerance.
        """
        k = kernel()
        # check that a very well-sampled case gives 1.0
        assert np.isclose(total_kernel_weight(k, 20), 1.0, rtol=1.0e-3)
        # check that the maximum size gives 1.0 within 1%
        assert np.isclose(
            total_kernel_weight(k, kernel.max_valid_size * 0.999), 1.0, rtol=1.0e-2
        )
        # check that with a larger size is more than 1% from 1.0
        if kernel not in (DiracDeltaKernel,):
            # DiracDeltaKernel will still converge to right answer, but is
            # a poor approximation on geometric grounds
            assert not np.isclose(
                total_kernel_weight(k, kernel.max_valid_size * 1.1), 1.0, rtol=1.0e-2
            )

    @pytest.mark.parametrize("truncate", (1.0, 2.0, 3.0, 4.0, 5.0, 6.0))
    def test_kernel_validation_Gaussian(self, truncate):
        """
        Check that our kernel integral approximations hold within the stated
        tolerance, in other words that we conserve mass within the same
        tolerance.
        """
        if truncate < 2.0:
            with pytest.raises(RuntimeError, match="with truncation <2sigma"):
                k = _GaussianKernel(truncate=truncate)
            return
        else:
            k = _GaussianKernel(truncate=truncate)
        # check that a very well-sampled case gives 1.0
        assert np.isclose(total_kernel_weight(k, 20), 1.0, rtol=3.0e-3)
        # check that the minimum size gives 1.0 within 1%
        assert np.isclose(
            total_kernel_weight(k, k.min_valid_size * 1.1), 1.0, rtol=1.0e-2
        )
        # don't check that with a smaller size is more than 1% from 1.0, off-centre
        # particle in pixel can fail even when this passes
        # assert not np.isclose(
        #     total_kernel_weight(k, k.min_valid_size * 0.9), 1.0, rtol=1.0e-2
        # )

    @pytest.mark.parametrize(
        ("kernel", "kernel_args"),
        (
            (_GaussianKernel, (2,)),
            (_GaussianKernel, (3,)),
            (_GaussianKernel, (4,)),
            (_GaussianKernel, (5,)),
            (_GaussianKernel, (6,)),
            (_WendlandC2Kernel, None),
            (_WendlandC6Kernel, None),
            (_CubicSplineKernel, None),
            (_QuarticSplineKernel, None),
            (DiracDeltaKernel, None),
        ),
    )
    def test_kernel_confirm_validation(self, kernel, kernel_args):
        """
        Setup a source at 3 Mpc and a datacube such that pixel scale is 1kpc. Provide
        smoothing lengths just above and below the minimum or maximum size and check that
        validation passes/fails accordingly.
        """
        if kernel_args is not None:

            def k():
                return kernel(*kernel_args)

        else:

            def k():
                return kernel()

        if hasattr(k(), "min_valid_size"):
            testcases = ((True, 0.9), (False, 1.1))
            threshold = k().min_valid_size
        elif hasattr(k(), "max_valid_size"):
            testcases = ((True, 1.1), (False, 0.9))
            threshold = k().max_valid_size
        for raises, factor in testcases:
            source = SPHSource(
                T_g=np.ones(1) * 1.0e4 * U.K,
                mHI_g=np.ones(1) * 1.0e4 * U.Msun,
                xyz_g=np.ones((1, 3)) * 1.0e-3 * U.kpc,
                vxyz_g=np.zeros((1, 3)) * U.km * U.s**-1,
                hsm_g=np.ones(1) * threshold * factor * U.kpc,
                distance=3 * U.Mpc,
            )
            datacube = DataCube(
                n_px_x=100,
                n_px_y=100,
                n_channels=32,
                px_size=68.75493542 * U.arcsec,
                channel_width=10 * U.km / U.s,
                spectral_centre=source.vsys,
            )
            m = Martini(
                source=source,
                datacube=datacube,
                beam=None,
                noise=None,
                spectral_model=GaussianSpectrum(),
                sph_kernel=k(),
                quiet=True,
            )
            if raises:
                with pytest.raises(RuntimeError, match="use this with care"):
                    m.sph_kernel._confirm_validation(quiet=True)
            else:
                m.sph_kernel._confirm_validation(quiet=True)  # should not raise


class TestAdaptiveKernels:
    @pytest.mark.parametrize(
        "kernel",
        (
            WendlandC2Kernel,
            WendlandC6Kernel,
            CubicSplineKernel,
            QuarticSplineKernel,
        ),
    )
    def test_kernel_selection(
        self, kernel, adaptive_kernel_test_source, adaptive_kernel_test_datacube
    ):
        """
        Test that the kernel is chosen correctly from those available.

        The smoothing lengths are respectively: (3.0, 1.0, 0.55, 0.1) kpc. Normally
        the first two should use the preferred kernel, the third should fall back to
        a `_GaussianKernel` with a large truncation radius, and the last should fall back
        to a `DiracDeltaKernel`. Assumes 1kpc pixels, which is what we'll use for testing.
        Note that the minimum valid size is compared to the smoothing length times the
        rescaling parameter, so for example
        `_WendlandC2Kernel.min_valid_size / _WendlandC2Kernel()._rescale` has a value of
        0.9477, so with our 1kpc pixels, a value in the smoothing length array of
        at least 0.9477kpc will use the preferred kernel, while less than 0.5kpc will
        use the `DiracDeltaKernel` and between these two will use the `_GaussianKernel`
        with truncation parameter of 6.
        """
        source = adaptive_kernel_test_source()
        source._init_skycoords()
        k = kernel()
        k._init_sm_lengths(source=source, datacube=adaptive_kernel_test_datacube)
        assert all(k.kernel_indices == np.array([0, 0, 2, 1]))

    def test_kernel_selection_Gaussian(
        self, adaptive_kernel_test_source, adaptive_kernel_test_datacube
    ):
        """
        Test that the kernel is chosen correctly from those available.

        The smoothing lengths are respectively (3.0, 1.0, 0.55, 0.1) kpc. For a
        `GaussianKernel` with truncation parameter of 3, the first should use the
        preferred kernel, the second and third should fall back to a `_GaussianKernel`
        with a truncation parameter of 6, and the last should fall back to a
        `DiracDeltaKernel`. The `_GaussianKernel` has a rescaling parameter of 1.
        """
        source = adaptive_kernel_test_source()
        source._init_skycoords()
        k = GaussianKernel(truncate=3.0)
        k._init_sm_lengths(source=source, datacube=adaptive_kernel_test_datacube)
        assert all(k.kernel_indices == np.array([0, 2, 2, 1]))

    @pytest.mark.parametrize("kernel", recommended_kernels)
    def test_raw_kernel_function(self, kernel):
        """
        Test that evaluating the `kernel` function gives the evaluation of the adaptive
        kernel's preferred kernel's `kernel` function.
        """
        eval_at = np.linspace(0, 3, 51)
        assert np.allclose(
            kernel().kernel(eval_at), kernel().kernels[0].kernel(eval_at)
        )

    @pytest.mark.parametrize("kernel", recommended_kernels)
    def test_eval_kernel_function(self, kernel):
        """
        Test that evaluating the `eval_kernel` function gives the evaluation of the
        adaptive kernel's preferred kernel's `eval_kernel` function.
        """
        eval_at = np.linspace(0, 3, 51)
        assert np.allclose(
            kernel().eval_kernel(eval_at, np.ones(eval_at.shape)),
            kernel().kernels[0].eval_kernel(eval_at, np.ones(eval_at.shape)),
        )



"""
Provides the :class:`~martini.sources.swiftgalaxy_source.SWIFTGalaxySource` class for
working with SWIFT simulations as input.
"""

import numpy as np
from .sph_source import SPHSource
from astropy import units as U
from astropy.coordinates import ICRS


class SWIFTGalaxySource(SPHSource):
    """
    Class abstracting HI sources designed to work with the :mod:`swiftsimio` and
    :mod:`swiftgalaxy` modules.

    Parameters
    ----------
    galaxy : ~swiftgalaxy.reader.SWIFTGalaxy
        Instance of a :class:`~swiftgalaxy.reader.SWIFTGalaxy`.

    distance : ~astropy.units.Quantity, optional
        :class:`~astropy.units.Quantity`, with dimensions of length.
        Source distance, also used to set the velocity offset via Hubble's law.
        (Default: ``3 * U.Mpc``)

    vpeculiar : ~astropy.units.Quantity, optional
        :class:`~astropy.units.Quantity`, with dimensions of velocity.
        Source peculiar velocity along the direction to the source centre.
        (Default: ``0 * U.km * U.s**-1``)

    rotation : dict, optional
        Must have a single key, which must be one of ``axis_angle``, ``rotmat`` or
        ``L_coords``. Note that the 'y-z' plane will be the one eventually placed in the
        plane of the "sky". The corresponding value must be:

        - ``axis_angle`` : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element a :class:`~astropy.units.Quantity` with \
        dimensions of angle, indicating the angle to rotate through.
        - ``rotmat`` : A (3, 3) :class:`~numpy.ndarray` specifying a rotation.
        - ``L_coords`` : A 2-tuple containing an inclination and an azimuthal \
        angle (both :class:`~astropy.units.Quantity` instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about \
        'y'). A 3-tuple may be provided instead, in which case the third \
        value specifies the position angle on the sky (second rotation about 'x'). \
        The default position angle is 270 degrees.

        (Default: ``np.eye(3)``)

    ra : ~astropy.units.Quantity, optional
        :class:`~astropy.units.Quantity`, with dimensions of angle.
        Right ascension for the source centroid. (Default: ``0 * U.deg``)

    dec : ~astropy.units.Quantity, optional
        :class:`~astropy.units.Quantity`, with dimensions of angle.
        Declination for the source centroid. (Default: ``0 * U.deg``)

    coordinate_frame : ~astropy.coordinates.builtin_frames.baseradec.BaseRADecFrame, \
    optional
        The coordinate frame assumed in converting particle coordinates to RA and Dec, and
        for transforming coordinates and velocities to the data cube frame. The frame
        needs to have a well-defined velocity as well as spatial origin. Recommended
        frames are :class:`~astropy.coordinates.GCRS`, :class:`~astropy.coordinates.ICRS`,
        :class:`~astropy.coordinates.HCRS`, :class:`~astropy.coordinates.LSRK`,
        :class:`~astropy.coordinates.LSRD` or :class:`~astropy.coordinates.LSR`. The frame
        should be passed initialized, e.g. ``ICRS()`` (not just ``ICRS``).
        (Default: ``astropy.coordinates.ICRS()``)
    """

    def __init__(
        self,
        galaxy,
        distance=3.0 * U.Mpc,
        vpeculiar=0 * U.km / U.s,
        rotation={"rotmat": np.eye(3)},
        ra=0.0 * U.deg,
        dec=0.0 * U.deg,
        coordinate_frame=ICRS(),
        _mHI_g=None,
    ):
        h = galaxy.metadata.cosmology.h
        mHI_g = (
            galaxy.gas.atomic_hydrogen_masses.to_astropy() if _mHI_g is None else _mHI_g
        )
        particles = dict(
            xyz_g=galaxy.gas.coordinates.to_astropy(),
            vxyz_g=galaxy.gas.velocities.to_astropy(),
            T_g=galaxy.gas.temperatures.to_astropy(),
            # SWIFT guarantees smoothing lengths are kernel FWHM, use directly:
            hsm_g=galaxy.gas.smoothing_lengths.to_astropy(),
            mHI_g=mHI_g,
        )
        super().__init__(
            distance=distance,
            vpeculiar=vpeculiar,
            rotation=rotation,
            ra=ra,
            dec=dec,
            h=h,
            coordinate_frame=coordinate_frame,
            coordinate_axis=1,
            **particles,
        )
        return



import os
from astropy.io import fits


class TestFITSCompliant:
    def test_cube_compliant(self, m):
        """
        Check that written cube complies with FITS standards enfored by astropy.
        """
        filename = "test_cube_compliant.fits"
        try:
            m.write_fits(filename, overwrite=True)
            with fits.open(filename) as hdul:
                hdul.verify("exception")  # errors here if failing
        finally:
            if os.path.exists(filename):
                os.remove(filename)

    def test_beam_compliant(self, m):
        """
        Check that written beam file complies with FITS standards enforced by astropy.
        """
        filename = "test_beam_compliant.fits"
        try:
            m.write_beam_fits(filename, overwrite=True)
            with fits.open(filename) as hdul:
                hdul.verify("exception")  # errors here if failing
        finally:
            if os.path.exists(filename):
                os.remove(filename)



from martini.__version__ import __version__


class TestVersion:
    def test_pyproject_version(self):
        """
        Check that pyproject version matches version file.
        """
        with open("pyproject.toml") as f:
            pyproject_content = f.read()
        assert 'version = "' + __version__ + '"' in pyproject_content

    def test_codemeta_version(self):
        """
        Check that codemeta version matches version file.
        """
        with open("codemeta.json") as f:
            codemeta_content = f.read()
        assert '"version": "' + __version__ + '",' in codemeta_content



"""
Provides classes to represent SPH kernels for smoothing particles onto the pixel grid.
"""

from abc import ABCMeta, abstractmethod
import numpy as np
import astropy.units as U
from scipy.special import erf
from scipy.optimize import fsolve


def find_fwhm(f):
    """
    Solve for the FWHM of the input function.

    Intended for SPH kernels and therefore assumes maximum occurs at :math:`f(0)` and
    symmetry around :math:`f(0)`.

    Parameters
    ----------
    f : callable
        Function for which to find the FWHM.

    Returns
    -------
    out : float
        FWHM of the input function.
    """
    return 2 * fsolve(lambda q: f(q) - f(np.zeros(1)) / 2, 0.5)[0]


class _BaseSPHKernel(object):
    """
    Abstract base class for classes implementing SPH kernels to inherit from.

    Classes inheriting from :class:`~martini.sph_kernels._BaseSPHKernel` must implement
    three methods: :meth:`~martini.sph_kernels._BaseSPHKernel.kernel`,
    :meth:`martini.sph_kernels._BaseSPHKernel._kernel_integral` and
    :meth:`~martini.sph_kernels._BaseSPHKernel_validate`.

    :meth:`~martini.sph_kernels._BaseSPHKernel.kernel` should define the kernel function,
    normalized such that its volume integral is ``1``.

    :meth:`~martini.sph_kernels._BaseSPHKernel._kernel_integral` should define the
    integral of the kernel over a pixel given the distance between the pixel centre and
    the particle centre, and the smoothing length (both in units of pixels). The integral
    should be normalized so that evaluated over the entire kernel it is equal to ``1``.

    :meth:`~martini.sph_kernels._BaseSPHKernel._validate` should check whether any
    approximations converge to sufficient accuracy (for instance, depending on the
    ratio of the pixel size and smoothing length), and raise an error if not. It should
    return a boolean array with ``True`` for particles which pass validation, and
    ``False`` otherwise.
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        self._rescale = 1
        return

    def _px_weight(self, dij, mask=None):
        """
        Calculate kernel integral using scaled smoothing lengths.

        This is the method that should be called by other modules in
        :mod:`martini`, rather than
        :meth:`~martini.sph_kernels._BaseSPHKernel._kernel_integral`.

        Parameters
        ----------
        dij : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Distances from pixel centre to particle positions, in pixels.

        mask : ~numpy.typing.ArrayLike or slice, optional
            Boolean mask to apply to any maskable attributes. (Default: ``None``)

        Returns
        -------
        out : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels^-2.
            Integral of smoothing kernel over pixel, per unit pixel area.
        """
        if mask is not None:
            try:
                rescale = self._rescale[mask]
            except (TypeError, IndexError):
                rescale = self._rescale
            rescaled_h = self.sm_lengths[mask] * rescale
        else:
            rescaled_h = self.sm_lengths * self._rescale
        return self._kernel_integral(dij, rescaled_h, mask=mask)

    def _confirm_validation(self, noraise=False, quiet=False):
        """
        Verify kernel accuracy using scaled smoothing lengths.

        This is the method that should be called by other modules in
        :mod:`martini`, rather than :meth:`~martini.sph_kernels._BaseSPHKernel._validate`.

        Parameters
        ----------
        noraise : bool
            If ``True``, don't raise error if validation fails. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """

        return self._validate(self.sm_lengths, noraise=noraise, quiet=quiet)

    def _validate_error(self, msg, sm_lengths, valid, noraise=False, quiet=False):
        """
        Function managing handling of kernel validation errors.

        Parameters
        ----------
        msg : str
            Error message.

        sm_lengths : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths in pixel units.

        valid : ~numpy.typing.ArrayLike
            Boolean array specifying which particles fail validation.

        noraise : bool
            If ``True``, don't raise error if validation fails. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """
        if not quiet:
            print(f"    ---------{self.__class__.__name__} VALIDATION---------")
            print("    Median smoothing length: ", np.median(sm_lengths), "px")
            print("    Minimum smoothing length: ", np.min(sm_lengths), "px")
            print("    Maximum smoothing length: ", np.max(sm_lengths), "px")
            print(
                "    Smoothing length histogram (np.histogram):",
                np.histogram(sm_lengths),
            )
            print(
                "    ",
                np.sum(np.logical_not(valid)),
                "/",
                sm_lengths.size,
                "smoothing lengths fail validation.",
            )
            print(
                "    -----------------------------" + "-" * len(self.__class__.__name__)
            )
        if not noraise:
            raise RuntimeError(msg)
        return

    def eval_kernel(self, r, h):
        """
        Evaluate the kernel, handling array casting and rescaling.

        The kernel parameter is defined:

        .. math::

            q = r / h

        Parameters
        ----------
        r : ~numpy.typing.ArrayLike or ~astropy.units.Quantity
            Distance parameter, same units (if applicable) as ``h``.
        h : ~numpy.typing.ArrayLike or ~astropy.units.Quantity
            Smoothing scale parameter (FWHM), same units (if applicable) as ``r``.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at position(s) ``r / h``.
        """

        q = np.array(r / h / self._rescale)
        if isinstance(q, U.Quantity):
            q = q.to_value(U.dimensionless_unscaled)
        scalar_input = q.ndim == 0
        W = self.kernel(q)
        W /= np.power(self._rescale, 3)
        if scalar_input:
            return W.item()
        else:
            return W

    def _apply_mask(self, mask):
        """
        Apply a mask to maskable attributes.

        Parameters
        ----------
        mask : ~numpy.typing.ArrayLike
            Boolean mask to apply to any maskable attributes.
        """
        self.sm_lengths = self.sm_lengths[mask]
        self.sm_ranges = self.sm_ranges[mask]

        return

    def _init_sm_lengths(self, source=None, datacube=None):
        """
        Determine kernel sizes in pixel units.

        Parameters
        ----------
        source : ~martini.sources.sph_source.SPHSource
            The source providing the kernel sizes.

        datacube : martini.datacube.DataCube
            The datacube providing the pixel scale.
        """
        self.sm_lengths = np.arctan(
            source.hsm_g
            / source.skycoords.transform_to(datacube.coordinate_frame).distance
        ).to(U.pix, U.pixel_scale(datacube.px_size / U.pix))

        return

    def _init_sm_ranges(self):
        """
        Determine maximum number of pixels reached by kernel.
        """
        # store as float: use case for np.inf in GlobalProfile:
        self.sm_ranges = np.ceil(self.sm_lengths * self.size_in_fwhm)

        return

    @abstractmethod
    def kernel(self, q):
        """
        Abstract method; evaluate the kernel.

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """
        pass

    @abstractmethod
    def _kernel_integral(self, dij, h, mask=np.s_[...]):
        """
        Abstract method; calculate the kernel integral over a pixel.

        Parameters
        ----------
        dij : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Distances from pixel centre to particle positions, in pixels.
        h : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths (FWHM), in pixels.
        mask : ~numpy.typing.ArrayLike or slice
            Boolean array, or slice. If the kernel has other internal properties to
            mask, it may use this.

        Returns
        -------
        out : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels^-2.
            Integral of smoothing kernel over pixel, per unit pixel area.
        """

        pass

    @abstractmethod
    def _validate(self, sm_lengths, noraise=False, quiet=False):
        """
        Abstract method; check conditions for validity of kernel integral
        calculation.

        Some approximations may only converge if the ratio of the pixel size
        and the smoothing length is sufficiently large, or sufficiently small.
        This method should check these conditions and raise errors when
        appropriate. The smoothing lengths are provided normalized to the pixel
        size. :class:`~martini.sph_kernels._AdaptiveKernel` needs to force errors
        not to raise, other classes should just define ``**kwargs`` in the signature.

        Parameters
        ----------
        sm_lengths : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths, in units of pixels.

        noraise : bool
            If ``True``, suppress exceptions. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """

        pass


class _WendlandC2Kernel(_BaseSPHKernel):
    """
    Implementation of the Wendland C2 kernel integral.

    The Wendland C2 kernel is used in the EAGLE code and derivatives (not in
    Gadget/Gadget2!). The exact integral is usually too slow to be practical;
    the implementation here approximates the kernel amplitude as constant
    across the pixel, which converges to within 1% of the exact integral
    provided the SPH smoothing lengths are at least 1.51 pixels in size.

    The expression:

    .. math::
        W(q) = \\begin{cases}
        \\frac{21}{2\\pi}(1-q)^4(4q+1)
        &{\\rm for}\\;0 \\leq q < 1\\\\
        0 &{\\rm for}\\;q \\geq 1
        \\end{cases}

    defines the WendlandC2 kernel.
    """

    min_valid_size = 1.51

    def __init__(self):
        super().__init__()
        _unscaled_fwhm = find_fwhm(lambda r: self.eval_kernel(r, 1))
        self.size_in_fwhm = 1 / _unscaled_fwhm
        self._rescale /= _unscaled_fwhm

        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The WendlandC2 kernel is here defined as:

        .. math::
            W(q) = \\begin{cases}
            \\frac{21}{2\\pi}(1-q)^4(4q+1)
            &{\\rm for}\\;0 \\leq q < 1\\\\
            0 &{\\rm for}\\;q \\geq 1
            \\end{cases}

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """

        W = np.where(q < 1, np.power(1 - q, 4) * (4 * q + 1), np.zeros(q.shape))
        W *= 21 / 2 / np.pi
        return W

    def _kernel_integral(self, dij, h, mask=np.s_[...]):
        """
        Calculate the kernel integral over a pixel.

        The formula used approximates the kernel amplitude as constant
        across the pixel area and converges to the true value within 1%
        for smoothing lengths >= 1.51 pixels.

        Parameters
        ----------
        dij : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Distances from pixel centre to particle positions, in pixels.
        h : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths (FWHM), in pixels.
        mask : ~numpy.typing.ArrayLike or slice
            Boolean array, or slice. If the kernel has other internal properties to mask,
            it may use this.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Approximate kernel integral over the pixel area.
        """

        dr2 = np.power(dij, 2).sum(axis=0)
        retval = np.zeros(h.shape)
        R2 = dr2 / (h * h)
        retval[R2 == 0] = 2.0 / 3.0
        use = np.logical_and(R2 < 1, R2 != 0)
        R2 = R2[use]
        A = np.sqrt(1 - R2)
        retval[use] = 5 * R2 * R2 * (0.5 * R2 + 3) * np.log(
            (1 + A) / np.sqrt(R2)
        ) + A * (-27.0 / 2.0 * R2 * R2 - 14.0 / 3.0 * R2 + 2.0 / 3.0)
        norm = 21 / 2 / np.pi
        return retval * norm / np.power(h, 2)

    def _validate(self, sm_lengths, noraise=False, quiet=False):
        """
        Check conditions for validity of kernel integral calculation.

        Parameters
        ----------
        sm_lengths : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths (FWHM), in units of pixels.

        noraise : bool
            If ``True``, suppress kernel validation errors. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """
        valid = sm_lengths >= self.min_valid_size * U.pix
        if np.logical_not(valid).any():
            self._validate_error(
                "martini.sph_kernels._WendlandC2Kernel._validate:\n"
                f"SPH smoothing lengths must be >= {self.min_valid_size:f} px in "
                "size for WendlandC2 kernel integral "
                "approximation accuracy within 1%.\nThis check "
                "may be disabled by calling "
                "martini.martini.Martini.insert_source_in_cube with "
                "'skip_validation=True', but use this with "
                "care.\n",
                sm_lengths,
                valid,
                noraise=noraise,
                quiet=quiet,
            )

        return valid


class _WendlandC6Kernel(_BaseSPHKernel):
    """
    Implementation of the Wendland C6 kernel integral.

    The Wendland C6 kernel is used in the Magneticum code (not in
    Gadget/Gadget2!). The exact integral is usually too slow to be practical;
    the implementation here approximates the kernel amplitude as constant
    across the pixel, which converges to within 1% of the exact integral
    provided the SPH smoothing lengths are at least 1.29 pixels in size.

    The expression:

    .. math::
        W(q) = \\begin{cases}
        \\frac{1365}{64 \\pi} (1 - q)^8 (1 + 8q + 25q^2 + 32q^3)
        &{\\rm for}\\;0 \\leq q < 1\\\\
        0 &{\\rm for}\\;q \\geq 1
        \\end{cases}

    defines the WendlandC6 kernel.
    """

    min_valid_size = 1.29

    def __init__(self):
        super().__init__()
        _unscaled_fwhm = find_fwhm(lambda r: self.eval_kernel(r, 1))
        self.size_in_fwhm = 1 / _unscaled_fwhm
        self._rescale /= _unscaled_fwhm
        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The WendlandC6 kernel is here defined as:

        .. math::
            W(q) = \\begin{cases}
            \\frac{1365}{64 \\pi} (1 - q)^8 (1 + 8q + 25q^2 + 32q^3)
            &{\\rm for}\\;0 \\leq q < 1\\\\
            0 &{\\rm for}\\;q \\geq 1
            \\end{cases}

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """

        W = np.where(
            q < 1,
            np.power(1 - q, 8)
            * (1 + 8 * q + 25 * np.power(q, 2) + 32 * np.power(q, 3)),
            np.zeros(q.shape),
        )
        W *= 1365 / 64 / np.pi
        return W

    def _kernel_integral(self, dij, h, mask=np.s_[...]):
        """
        Calculate the kernel integral over a pixel.

        The formula used approximates the kernel amplitude as constant
        across the pixel area and converges to the true value within 1%
        for smoothing lengths >= 1.29 pixels.

        Parameters
        ----------
        dij : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Distances from pixel centre to particle positions, in pixels.
        h : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths, in pixels.
        mask : ~numpy.typing.ArrayLike or slice
            Boolean array, or slice. If the kernel has other internal properties to mask,
            it may use this.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Approximate kernel integral over the pixel area.
        """

        def indef(R, z):
            """
            Evaluate expression involved in the integral.

            Parameters
            ----------
            R : ~numpy.typing.ArrayLike
                Dimensionless cylindrical radii.

            z : ~numpy.typing.ArrayLike
                Dimensionless vertical coordinates.

            Returns
            -------
            out : ~numpy.typing.ArrayLike
                Result of the expression.
            """

            return (
                -231 * np.power(R, 10) * z
                - 385 * np.power(R, 8) * np.power(z, 3)
                - 1155 * np.power(R, 8) * z
                - 462 * np.power(R, 6) * np.power(z, 5)
                - 1540 * np.power(R, 6) * np.power(z, 3)
                - 462 * np.power(R, 6) * z
                - 330 * np.power(R, 4) * np.power(z, 7)
                - 1386 * np.power(R, 4) * np.power(z, 5)
                - 462 * np.power(R, 4) * np.power(z, 3)
                + 66 * np.power(R, 4) * z
                - (128 + 1 / 3) * np.power(R, 2) * np.power(z, 9)
                - 660 * np.power(R, 2) * np.power(z, 7)
                - 277.2 * np.power(R, 2) * np.power(z, 5)
                + 44 * np.power(R, 2) * np.power(z, 3)
                + 8 / 3 * np.power(z, 11) * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + (16 + 4 / 15)
                * np.power(R, 2)
                * np.power(z, 9)
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 70.4 * np.power(z, 9) * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 360.8
                * np.power(R, 2)
                * np.power(z, 7)
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 132 * np.power(z, 7) * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 550
                * np.power(R, 2)
                * np.power(z, 5)
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                - 11 * np.power(R, 2) * z
                + 7.21875
                * np.power(R, 12)
                * (np.log(np.sqrt(np.power(R, 2) + np.power(z, 2)) + z))
                + 24.7813
                * np.power(R, 10)
                * z
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 173.25
                * np.power(R, 10)
                * (np.log(np.sqrt(np.power(R, 2) + np.power(z, 2)) + z))
                + 530.75
                * np.power(R, 8)
                * z
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 288.75
                * np.power(R, 8)
                * (np.log(np.sqrt(np.power(R, 2) + np.power(z, 2)) + z))
                + 47.4792
                * np.power(R, 8)
                * np.power(z, 3)
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 767.25
                * np.power(R, 6)
                * z
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 58.0167
                * np.power(R, 6)
                * np.power(z, 5)
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 819.5
                * np.power(R, 6)
                * np.power(z, 3)
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 41.7
                * np.power(R, 4)
                * np.power(z, 7)
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 752.4
                * np.power(R, 4)
                * np.power(z, 5)
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                + 896.5
                * np.power(R, 4)
                * np.power(z, 3)
                * (np.sqrt(np.power(R, 2) + np.power(z, 2)))
                - 21 * np.power(z, 11)
                - (128 + 1 / 3) * np.power(z, 9)
                - 66 * np.power(z, 7)
                + 13.2 * np.power(z, 5)
                - 11 / 3 * np.power(z, 3)
                + z
            )

        dr2 = np.power(dij, 2).sum(axis=0)
        retval = np.zeros(h.shape)
        R = np.sqrt(dr2) / h
        use = np.logical_and(R < 1, R != 0)
        norm = 1365 / 64 / np.pi
        zmax = np.sqrt(1 - np.power(R[use], 2))
        retval[use] = norm * 2 * (indef(R[use], zmax) - indef(R[use], 0))
        retval[R == 0] = norm * 2 * (4 / 15)
        retval = retval / np.power(h, 2)
        return retval

    def _validate(self, sm_lengths, noraise=False, quiet=False):
        """
        Check conditions for validity of kernel integral calculation.

        Parameters
        ----------
        sm_lengths : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths, in units of pixels.

        noraise : bool
            If ``True``, suppress kernel validation errors. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """

        valid = sm_lengths >= self.min_valid_size * U.pix
        if np.logical_not(valid).any():
            self._validate_error(
                "martini.sph_kernels._WendlandC6Kernel._validate:\n"
                f"SPH smoothing lengths must be >= {self.min_valid_size:f} px in "
                "size for WendlandC6 kernel integral "
                "approximation accuracy within 1%.\nThis check "
                "may be disabled by calling "
                "martini.martini.Martini.insert_source_in_cube with "
                "'skip_validation=True', but use this with "
                "care.",
                sm_lengths,
                valid,
                noraise=noraise,
                quiet=quiet,
            )
        return valid


class _CubicSplineKernel(_BaseSPHKernel):
    """
    Implementation of the cubic spline (M4) kernel integral.

    The cubic spline is the 'classic' SPH kernel. The exact integral is usually
    too slow to be practical; the implementation here approximates the kernel
    amplitude as constant across the pixel, which converges to within 1% of
    the exact integral provided the SPH smoothing lengths are at least 1.16
    pixels in size.

    The expression:

    .. math ::
        W(q) = \\frac{8}{\\pi}\\begin{cases}
        (1 - 6q^2(1 - q))
        &{\\rm for}\\;0 \\leq q < \\frac{1}{2}\\\\
        2(1 - q)^3
        &{\\rm for}\\;\\frac{1}{2} \\leq q < 1\\\\
        0
        &{\\rm for}\\;q \\geq 1
        \\end{cases}

    defines the cubic spline kernel.
    """

    min_valid_size = 1.16

    def __init__(self):
        super().__init__()
        _unscaled_fwhm = find_fwhm(lambda r: self.eval_kernel(r, 1))
        self.size_in_fwhm = 1 / _unscaled_fwhm
        self._rescale /= _unscaled_fwhm
        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The cubic spline kernel is here defined as:

        .. math ::
            W(q) = \\frac{8}{\\pi}\\begin{cases}
            (1 - 6q^2(1 - q))
            &{\\rm for}\\;0 \\leq q < \\frac{1}{2}\\\\
            2(1 - q)^3
            &{\\rm for}\\;\\frac{1}{2} \\leq q < 1\\\\
            0
            &{\\rm for}\\;q \\geq 1
            \\end{cases}

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """

        W = np.where(
            q < 0.5, 1 - 6 * np.power(q, 2) + 6 * np.power(q, 3), 2 * np.power(1 - q, 3)
        )
        W[q > 1] = 0
        W *= 8 / np.pi
        return W

    def _kernel_integral(self, dij, h, mask=np.s_[...]):
        """
        Calculate the kernel integral over a pixel.

        The formula used approximates the kernel amplitude as constant across
        the pixel area and converges to the true value within 1% for smoothing
        lengths >= 1.16 pixels.

        Parameters
        ----------
        dij : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.
        h : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels
            Particle smoothing lengths, in pixels.
        mask : ~numpy.typing.ArrayLike or slice
            Boolean array, or slice. If the kernel has other internal properties to mask,
            it may use this.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Approximate kernel integral over the pixel area.
        """

        dij *= 2  # changes interval from [0, 2) to [0, 1)
        dr2 = np.power(dij, 2).sum(axis=0)
        retval = np.zeros(h.shape)
        R2 = dr2 / (h * h)
        retval[R2 == 0] = 11.0 / 16.0 + 0.25 * 0.25
        case1 = np.logical_and(R2 > 0, R2 <= 1)
        case2 = np.logical_and(R2 > 1, R2 <= 4)

        R2_1 = R2[case1]
        R2_2 = R2[case2]
        A_1 = np.sqrt(1 - R2_1)
        B_1 = np.sqrt(4 - R2_1)
        B_2 = np.sqrt(4 - R2_2)
        I1 = (
            A_1
            - 0.5 * np.power(A_1, 3)
            - 1.5 * R2_1 * A_1
            + 3.0 / 32.0 * A_1 * (3 * R2_1 + 2)
            + 9.0 / 32.0 * R2_1 * R2_1 * (np.log(1 + A_1) - np.log(np.sqrt(R2_1)))
        )
        I2 = (
            -B_2 * (3 * R2_2 + 56) / 4.0
            - 3.0 / 8.0 * R2_2 * (R2_2 + 16) * np.log((2 + B_2) / np.sqrt(R2_2))
            + 2 * (3 * R2_2 + 4) * B_2
            + 2 * np.power(B_2, 3)
        )
        I3 = (
            -B_1 * (3 * R2_1 + 56) / 4.0
            + A_1 * (4 * R2_1 + 50) / 8.0
            - 3.0 / 8.0 * R2_1 * (R2_1 + 16) * np.log((2 + B_1) / (1 + A_1))
            + 2 * (3 * R2_1 + 4) * (B_1 - A_1)
            + 2 * (np.power(B_1, 3) - np.power(A_1, 3))
        )
        retval[case1] = I1 + 0.25 * I3
        retval[case2] = 0.25 * I2
        # 1.597 is normalization s.t. kernel integral = 1 for particle mass = 1
        # rescaling from interval [0, 2) to [0, 1) requires mult. by 4
        return retval / 1.59689476201133 / np.power(h, 2) * 4

    def _validate(self, sm_lengths, noraise=False, quiet=False):
        """
        Check conditions for validity of kernel integral calculation.

        Parameters
        ----------
        sm_lengths : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths, in units of pixels.

        noraise : bool
            If ``True``, suppress kernel validation errors. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """

        valid = sm_lengths >= self.min_valid_size * U.pix
        if np.logical_not(valid).any():
            self._validate_error(
                "martini.sph_kernels._CubicSplineKernel._validate:\n"
                f"SPH smoothing lengths must be >= {self.min_valid_size:f} px in "
                "size for CubicSplineKernel kernel integral "
                "approximation accuracy within 1%.\nThis check "
                "may be disabled by calling "
                "martini.martini.Martini.insert_source_in_cube with "
                "'skip_validation=True', but use this with "
                "care.",
                sm_lengths,
                valid,
                noraise=noraise,
                quiet=quiet,
            )
        return valid


class _GaussianKernel(_BaseSPHKernel):
    """
    Implementation of a (truncated) Gaussian kernel integral.

    Calculates the kernel integral over a pixel. The 3 integrals (along dx,
    dy, dz) are evaluated exactly, however the truncation is implemented
    approximately, erring on the side of integrating slightly further than
    the truncation radius.

    The Gaussian kernel is here defined as:

    .. math::
        W(q) = \\begin{cases}
        (\\sqrt{2\\pi}\\sigma)^{-3}
        \\exp\\left(-\\frac{1}{2}\\left(\\frac{q}{\\sigma}\\right)^2\\right)
        &{\\rm for}\\;0 \\leq q < t\\\\
        0 &{\\rm for}\\;q > t
        \\end{cases}

    with :math:`\\sigma=(2\\sqrt{2\\log(2)})^{-1}`, s.t. FWHM = 1, and
    :math:`t` being the truncation radius.

    Parameters
    ----------
    truncate : float, optional
        Number of standard deviations at which to truncate kernel.
        Truncation radii <2 would lead to large errors and are not permitted.
        (Default: ``3``)
    """

    def __init__(self, truncate=3.0):
        self.truncate = truncate
        if self.truncate < 2:
            raise RuntimeError(
                "GaussianKernel with truncation <2sigma will "
                "cause large errors in total mass."
            )
        elif (self.truncate >= 2) and (self.truncate < 3):
            self.min_valid_size = 3.7
        elif (self.truncate >= 3) and (self.truncate < 4):
            self.min_valid_size = 2.3357
        elif (self.truncate >= 4) and (self.truncate < 5):
            self.min_valid_size = 1.1288
        elif (self.truncate >= 5) and (self.truncate < 6):
            self.min_valid_size = 0.45
        elif self.truncate >= 6:
            self.min_valid_size = 0.336

        self.norm = erf(self.truncate / np.sqrt(2)) - 2 * self.truncate / np.sqrt(
            2 * np.pi
        ) * np.exp(-np.power(self.truncate, 2) / 2)
        super().__init__()
        self.size_in_fwhm = self.truncate / (2 * np.sqrt(2 * np.log(2)))
        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The Gaussian kernel is here defined as:

        .. math::
            W(q) = \\begin{cases}
            (\\sqrt{2\\pi}\\sigma)^{-3}
            \\exp\\left(-\\frac{1}{2}\\left(\\frac{q}{\\sigma}\\right)^2\\right)
            &{\\rm for}\\;0 \\leq q < t\\\\
            0 &{\\rm for}\\;q > t
            \\end{cases}

        with :math:`\\sigma=(2\\sqrt{2\\log(2)})^{-1}`, s.t. FWHM = 1, and
        :math:`t` being the truncation radius.

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """

        sig = 1 / (2 * np.sqrt(2 * np.log(2)))  # s.t. FWHM = 1
        return (
            np.where(
                q < self.truncate * sig,
                np.power(sig * np.sqrt(2 * np.pi), -3)
                * np.exp(-np.power(q / sig, 2) / 2),
                np.zeros(q.shape),
            )
            / self.norm
        )

    def _kernel_integral(self, dij, h, mask=np.s_[...]):
        """
        Calculate the kernel integral over a pixel.

        The 3 integrals (along :math:`dx`, :math:`dy`, :math:`dz`) are evaluated exactly,
        however the truncation is implemented approximately, erring on the side of
        integrating slightly further than the truncation radius.

        Parameters
        ----------
        dij : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Distances from pixel centre to particle positions, in pixels.
        h : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths (FWHM), in pixels.
        mask : ~numpy.typing.ArrayLike or slice
            Boolean array, or slice. If the kernel has other internal properties to mask,
            it may use this.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel integral over the pixel area.
        """
        sig = 1 / (2 * np.sqrt(2 * np.log(2)))  # s.t. FWHM = 1
        dr = np.sqrt(np.power(dij, 2).sum(axis=0))
        with np.errstate(invalid="ignore"):
            zmax = np.sqrt(np.power(self.truncate, 2) - np.power(dr / h / sig, 2))
        zmax = np.where(self.truncate > dr / h / sig, zmax, 0)
        x0 = (dij[0] - 0.5 * U.pix) / h / np.sqrt(2) / sig
        x1 = (dij[0] + 0.5 * U.pix) / h / np.sqrt(2) / sig
        y0 = (dij[1] - 0.5 * U.pix) / h / np.sqrt(2) / sig
        y1 = (dij[1] + 0.5 * U.pix) / h / np.sqrt(2) / sig

        retval = (
            0.25 * erf(zmax / np.sqrt(2)) * (erf(x1) - erf(x0)) * (erf(y1) - erf(y0))
        )

        # explicit truncation not required as only pixels inside
        # truncation radius should be passed, next line useful for
        # testing, however
        retval[(dr - np.sqrt(0.5) * U.pix) / h / sig > self.truncate] = 0

        retval /= self.norm
        return retval * h.unit**-2

    def _validate(self, sm_lengths, noraise=False, quiet=False):
        """
        Check conditions for validity of kernel integral calculation.

        Parameters
        ----------
        sm_lengths : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths (FWHM), in units of pixels.

        noraise : bool
            If ``True``, suppress kernel validation errors. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """
        valid = sm_lengths >= self.min_valid_size * U.pix
        if np.logical_not(valid).any():
            self._validate_error(
                "martini.sph_kernels._GaussianKernel._validate:\n"
                f"SPH smoothing lengths must be >= {self.min_valid_size:f} px in "
                "size for GaussianKernel kernel integral "
                "approximation accuracy within 1%.\nThis check "
                "may be disabled by calling "
                "martini.martini.Martini.insert_source_in_cube with "
                "'skip_validation=True', but use this with "
                "care. Note that the minimum size depends on the kernel truncation.",
                sm_lengths,
                valid,
                noraise=noraise,
                quiet=quiet,
            )
        return valid


class DiracDeltaKernel(_BaseSPHKernel):
    """
    Implementation of a Dirac-delta kernel integral.

    The Dirac-delta kernel is here defined as:

    .. math::
        W(q) = \\begin{cases}
        \\infty &{\\rm for}\\;q = 0\\\\
        0 &{\\rm for}\\;q > 0
        \\end{cases}

    Parameters
    ----------
    size_in_fwhm : float
        In principle the width of a Dirac-delta kernel is 0, but this would
        lead to no particles contributing to any pixels. Ideally this would
        be set to approximately the size of the pixel, but setting it to
        the smoothing length (1.0) is acceptable given the validation condition.
        (Default: 1.0)
    """

    max_valid_size = 0.5

    def __init__(self, size_in_fwhm=1.0):
        super().__init__()
        self.size_in_fwhm = size_in_fwhm
        self._rescale = 1  # need this to be present
        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The Dirac-delta kernel is here defined as:

        .. math::
            W(q) = \\begin{cases}
            \\infty &{\\rm for}\\;q = 0\\\\
            0 &{\\rm for}\\;q > 0
            \\end{cases}

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """

        return np.where(q, np.zeros(q.shape), np.inf * np.ones(q.shape))

    def _kernel_integral(self, dij, h, mask=np.s_[...]):
        """
        Calculate the kernel integral over a pixel.

        The particles are approximated as point-like, ignoring any finite-sized
        kernel. This is a reasonable approximation provided the smoothing
        length is < 0.5 pixel in size, ideally << 1 pixel in size.

        Parameters
        ----------
        dij : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Distances from pixel centre to particle positions, in pixels.
        h : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths (FWHM), in pixels.
        mask : ~numpy.typing.ArrayLike or slice, optional
            Boolean mask to apply to any maskable attributes. (Default: ``np.s_[...]``)

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel integral over the pixel area.
        """
        return np.where((np.abs(dij) < 0.5 * U.pix).all(axis=0), 1, 0) * U.pix**-2

    def _validate(self, sm_lengths, noraise=False, quiet=False):
        """
        Check conditions for validity of kernel integral calculation.

        The Dirac-delta model approaches the exact integral when the smoothing
        length is << 1 pixel in size; at a minimum the smoothing length should
        be less than half the pixel size.

        Parameters
        ----------
        sm_lengths : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths (FWHM), in units of pixels.

        noraise : bool
            If ``True``, suppress kernel validation errors. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """

        valid = sm_lengths <= self.max_valid_size * U.pix
        if np.logical_not(valid).any():
            self._validate_error(
                "martini.sph_kernels.DiracDeltaKernel._validate:\n"
                f"provided smoothing scale (FWHM) must be <= {self.max_valid_size:f} "
                "px in size for DiracDelta kernel to be a "
                "reasonable approximation. Call "
                "martini.martini.Martini.insert_source_in_cube with "
                "'skip_validation=True' to override at the "
                "cost of accuracy, but use this with care.",
                sm_lengths,
                valid,
                noraise=noraise,
                quiet=quiet,
            )
        return valid


class _AdaptiveKernel(_BaseSPHKernel):
    """
    Allows use of multiple kernels to adapt to sph kernel-to-pixel size ratio.

    Other provided kernels generally use approximations which break down if
    the ratio of the pixel size and the sph smoothing length are above or below
    some threshold. This (meta-)kernel accepts a list of other kernels in order
    of decreasing priority. The validity of the approximations used in each
    will be checked in turn and the first usable kernel for a given particle
    will be used to smooth the particle onto the pixel grid. Note that the
    initialized source and datacube instances are required as the smoothing
    lengths and pixel sizes must be known at initialization of the
    :class:`~martini.sph_kernels._AdaptiveKernel` class. Note that if ``skip_validation``
    is used, any particles with no valid kernel will default to the first kernel in the
    list.

    Parameters
    ----------
    kernels : iterable
        An iterable containing classes inheriting from
        :class:`~martini.sph_kernels._BaseSPHKernel`.
        Kernels to use, ordered by decreasing priority.
    """

    def __init__(self, kernels):
        self.kernels = kernels
        super().__init__()
        self.size_in_fwhm = None  # initialized during Martini.__init__
        self._rescale = None  # initialized during Martini.__init__
        return

    def _init_sm_lengths(self, source=None, datacube=None):
        """
        Determine kernel sizes in pixel units.

        Parameters
        ----------
        source : ~martini.sources.sph_source.SPHSource
            The source providing the kernel sizes.

        datacube : martini.datacube.DataCube
            The datacube providing the pixel scale.
        """

        super()._init_sm_lengths(source=source, datacube=datacube)
        self.kernel_indices = -1 * np.ones(source.hsm_g.shape, dtype=int)
        for ik, K in enumerate(self.kernels):
            # if valid and not already assigned an earlier kernel, assign
            self.kernel_indices[
                np.logical_and(
                    self.kernel_indices == -1,
                    K._validate(self.sm_lengths * K._rescale, noraise=True, quiet=True),
                )
            ] = ik
        _sizes_in_fwhm = np.array([K.size_in_fwhm for K in self.kernels])
        self.size_in_fwhm = _sizes_in_fwhm[self.kernel_indices]
        # ensure default is 0th entry
        self.size_in_fwhm[self.kernel_indices == -1] = _sizes_in_fwhm[0]
        _rescales = np.array([K._rescale for K in self.kernels])
        self._rescale = _rescales[self.kernel_indices]
        # ensure default is 0th entry
        self._rescale[self.kernel_indices == -1] = _rescales[0]

        return

    def _apply_mask(self, mask):
        """
        Apply a mask to maskable attributes.

        Parameters
        ----------
        mask : ~numpy.typing.ArrayLike
            Boolean mask to apply to any maskable attributes.
        """
        self.size_in_fwhm = self.size_in_fwhm[mask]
        self._rescale = self._rescale[mask]
        self.kernel_indices = self.kernel_indices[mask]
        super()._apply_mask(mask)

        return

    def eval_kernel(self, r, h):
        """
        Evaluate the kernel, handling array casting and rescaling.

        Parameters
        ----------
        r : ~numpy.typing.ArrayLike or ~astropy.units.Quantity
            Distance parameter, same units as ``h``.
        h : ~numpy.typing.ArrayLike or ~astropy.units.Quantity
            Smoothing scale parameter (FWHM), same units as ``r``.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at position(s) ``r / h``.
        """

        return self.kernels[0].eval_kernel(r, h)

    def kernel(self, q):
        """
        Evaluate the kernel function of the preferred kernel.

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """
        return self.kernels[0].kernel(q)

    def _kernel_integral(self, dij, h, mask=np.s_[...]):
        """
        Calculate the kernel integral over a pixel.

        Adaptively determines which kernel to use.

        Parameters
        ----------
        dij : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Distances from pixel centre to particle positions, in pixels.
        h : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths (FWHM), in pixels.
        mask : ~numpy.typing.ArrayLike or slice
            Boolean array, or slice. If the kernel has other internal properties to mask,
            it may use this.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Approximate kernel integral over the pixel area.
        """

        retval = np.zeros(h.shape) * h.unit**-2
        for ik in np.unique(self.kernel_indices[mask]):
            K = self.kernels[0] if ik == -1 else self.kernels[ik]
            kmask = self.kernel_indices[mask] == ik
            retval[kmask] = K._kernel_integral(dij[:, kmask], h[kmask])
        return retval

    def _validate(self, sm_lengths, noraise=False, quiet=False):
        """
        Check conditions for validity of kernel integral calculation.

        Parameters
        ----------
        sm_lengths : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths (FWHM), in units of pixels.

        noraise : bool
            If ``True``, suppress kernel validation errors. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """

        valid = self.kernel_indices >= 0
        if np.logical_not(valid).any():
            self._validate_error(
                "martini.sph_kernels._AdaptiveKernel._validate:\n"
                "Some particles have no kernel candidate for which "
                "accuracy passes validation.\nThis check "
                "may be disabled by calling "
                "martini.martini.Martini.insert_source_in_cube with "
                "'skip_validation=True', but use this with "
                "care.\n",
                sm_lengths,
                valid,
                noraise=noraise,
                quiet=quiet,
            )

        return


class _QuarticSplineKernel(_BaseSPHKernel):
    """
    Implementation of the quartic spline kernel integral.

    The quartic spline (M5) kernel is used in the SPHENIX scheme (e.g. in Colibre). The
    exact integral is usually too slow to be practical; the implementation here
    approximates the kernel amplitude as constant across the pixel, which converges to
    within 1% of the exact integral provided the SPH smoothing lengths are at least 1.2385
    pixels in size.

    The expression:

    .. math ::
        W(q) = \\frac{15625}{512\\pi}\\begin{cases}
        (1 - q)^4 - 5(\\frac{3}{5} - q)^4 + 10(\\frac{1}{5}-q)^4
        &{\\rm for}\\;0 \\leq q < \\frac{1}{5}\\\\
        (1 - q)^4 - 5(\\frac{3}{5} - q)^4
        &{\\rm for}\\;\\frac{1}{5} \\leq q < \\frac{3}{5}\\\\
        (1 - q)^4
        &{\\rm for}\\;\\frac{3}{5} \\leq q < 1\\\\
        0
        &{\\rm for}\\;q \\geq 1
        \\end{cases}

    defines the quartic spline kernel.
    """

    min_valid_size = 1.2385

    def __init__(self):
        super().__init__()
        _unscaled_fwhm = find_fwhm(lambda r: self.eval_kernel(r, 1))
        self.size_in_fwhm = 1 / _unscaled_fwhm
        self._rescale /= _unscaled_fwhm
        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The quartic spline kernel is here defined as:

        .. math ::
            W(q) = \\frac{15625}{512\\pi}\\begin{cases}
            (1 - q)^4 - 5(\\frac{3}{5} - q)^4 + 10(\\frac{1}{5}-q)^4
            &{\\rm for}\\;0 \\leq q < \\frac{1}{5}\\\\
            (1 - q)^4 - 5(\\frac{3}{5} - q)^4
            &{\\rm for}\\;\\frac{1}{5} \\leq q < \\frac{3}{5}\\\\
            (1 - q)^4
            &{\\rm for}\\;\\frac{3}{5} \\leq q < 1\\\\
            0
            &{\\rm for}\\;q \\geq 1
            \\end{cases}

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """

        W = np.zeros(q.shape)
        mask1 = q < 0.2
        W[mask1] = (
            np.power(1 - q[mask1], 4)
            - 5 * np.power(0.6 - q[mask1], 4)
            + 10 * np.power(0.2 - q[mask1], 4)
        )
        mask2 = np.logical_and(q >= 0.2, q < 0.6)
        W[mask2] = np.power(1 - q[mask2], 4) - 5 * np.power(0.6 - q[mask2], 4)
        mask3 = np.logical_and(q >= 0.6, q < 1)
        W[mask3] = np.power(1 - q[mask3], 4)
        W *= 15625 / 512 / np.pi
        return W

    def _kernel_integral(self, dij, h, mask=np.s_[...]):
        """
        Calculate the kernel integral over a pixel.

        The formula used approximates the kernel amplitude as constant across
        the pixel area and converges to the true value within 1% for smoothing
        lengths >= 1.2385 pixels.

        Parameters
        ----------
        dij : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Distances from pixel centre to particle positions, in pixels.
        h : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths, in pixels.
        mask : ~numpy.typing.ArrayLike or slice
            Boolean array, or slice. If the kernel has other internal properties to mask,
            it may use this.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Approximate kernel integral over the pixel area.
        """

        dr = np.sqrt(np.power(dij, 2).sum(axis=0))
        retval = np.zeros(h.shape)
        R = (dr / h).to_value(U.dimensionless_unscaled)

        def IA(R, z, A):
            """
            Evaluate expression involved in the integral.

            Parameters
            ----------
            R : ~numpy.typing.ArrayLike
                Dimensionless cylindrical radii.

            z : ~numpy.typing.ArrayLike
                Dimensionless vertical coordinates.

            A : ~numpy.typing.ArrayLike
                Dimensionless amplitudes.

            Returns
            -------
            out : ~numpy.typing.ArrayLike
                Result of the expression.
            """
            q = np.sqrt(np.power(z, 2) + np.power(R, 2))
            R2 = np.power(R, 2)
            return (
                np.power(A, 4) * z
                - 2 * np.power(A, 3) * z * q
                + 2 * np.power(A, 2) * z * (3 * R2 + np.power(z, 2))
                - A * R2 * (4 * np.power(A, 2) + 3 * R2) * np.arcsinh(z / R) / 2
                - A * z * q * (5 * R2 + 2 * np.power(z, 2)) / 2
                + R2 * R2 * z
                + 2 * R2 * np.power(z, 3) / 3
                + np.power(z, 5) / 5
            )

        m1 = np.logical_and(R > 0, R < 0.2)
        retval[m1] += 10 * IA(R[m1], np.sqrt(0.2**2 - np.power(R[m1], 2)), 0.2)
        m2 = np.logical_and(R > 0, R < 0.6)
        retval[m2] -= 5 * IA(R[m2], np.sqrt(0.6**2 - np.power(R[m2], 2)), 0.6)
        m3 = np.logical_and(R > 0, R < 1.0)
        retval[m3] += IA(R[m3], np.sqrt(1 - np.power(R[m3], 2)), 1.0)
        retval[R == 0] = 384 / 3125
        # factor of 2 is because all integrals above are half-intervals
        retval *= 2 * 15625 / 512 / np.pi
        return retval / np.power(h, 2)

    def _validate(self, sm_lengths, noraise=False, quiet=False):
        """
        Check conditions for validity of kernel integral calculation.

        Parameters
        ----------
        sm_lengths : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of pixels.
            Particle smoothing lengths, in units of pixels.

        noraise : bool
            If ``True``, suppress kernel validation errors. (Default: ``False``)

        quiet : bool
            If ``True``, suppress reports on smoothing lengths. (Default: ``False``)
        """

        valid = sm_lengths >= self.min_valid_size * U.pix
        if np.logical_not(valid).any():
            self._validate_error(
                "martini.sph_kernels._QuarticSplineKernel._validate:\n"
                "SPH smoothing lengths must be >= {self.min_valid_size:f} px in "
                "size for QuarticSplineKernel kernel integral "
                "approximation accuracy within 1%.\nThis check "
                "may be disabled by calling "
                "martini.martini.Martini.insert_source_in_cube with "
                "'skip_validation=True', but use this with "
                "care.",
                sm_lengths,
                valid,
                noraise=noraise,
                quiet=quiet,
            )
        return valid


class WendlandC2Kernel(_AdaptiveKernel):
    """
    Implementation of the Wendland C2 kernel integral.

    The Wendland C2 kernel is used in the EAGLE code and derivatives (not in
    Gadget/Gadget2!). The exact integral is usually too slow to be practical;
    the implementation here approximates the kernel amplitude as constant
    across the pixel, which converges to within 1% of the exact integral
    provided the SPH smoothing lengths are at least 1.51 pixels in size.

    The WendlandC2 kernel is here defined as:

    .. math::
        W(q) = \\begin{cases}
        \\frac{21}{2\\pi}(1-q)^4(4q+1)
        &{\\rm for}\\;0 \\leq q < 1\\\\
        0 &{\\rm for}\\;q \\geq 1
        \\end{cases}

    This class falls back to the :class:`~martini.sph_kernels.DiracDeltaKernel` and
    :class:`~martini.sph_kernels.GaussianKernel` when the approximation used for the
    surface integral of the WendlandC2 kernel is not accurate to within better than 1%.
    To strictly use the WendlandC2 kernel, use
    :class:`~martini.sph_kernels._WendlandC2Kernel` instead.
    """

    def __init__(self):
        super().__init__(
            (_WendlandC2Kernel(), DiracDeltaKernel(), _GaussianKernel(truncate=6.0))
        )
        self.size_in_fwhm = None  # initialized during Martini.__init__
        self._rescale = None  # initialized during Martini.__init__
        return

    def kernel(self, q):
        """
        Evaluate the kernel function of the WendlandC2 kernel.

        The WendlandC2 kernel is here defined as:

        .. math::
            W(q) = \\begin{cases}
            \\frac{21}{2\\pi}(1-q)^4(4q+1)
            &{\\rm for}\\;0 \\leq q < 1\\\\
            0 &{\\rm for}\\;q \\geq 1
            \\end{cases}

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """
        return super().kernel(q)


class WendlandC6Kernel(_AdaptiveKernel):
    """
    Implementation of the Wendland C6 kernel integral.

    The Wendland C6 kernel is used in the Magneticum code (not in
    Gadget/Gadget2!). The exact integral is usually too slow to be practical;
    the implementation here approximates the kernel amplitude as constant
    across the pixel, which converges to within 1% of the exact integral
    provided the SPH smoothing lengths are at least 1.29 pixels in size.

    The WendlandC6 kernel is here defined as:

    .. math::
        W(q) = \\begin{cases}
        \\frac{1365}{64 \\pi} (1 - q)^8 (1 + 8q + 25q^2 + 32q^3)
        &{\\rm for}\\;0 \\leq q < 1\\\\
        0 &{\\rm for}\\;q \\geq 1
        \\end{cases}

    This class falls back to the :class:`~martini.sph_kernels.DiracDeltaKernel` and
    :class:`~martini.sph_kernels.GaussianKernel` when the approximation used for the
    surface integral of the WendlandC6 kernel is not accurate to within better than 1%.
    To strictly use the WendlandC6 kernel, use
    :class:`~martini.sph_kernels._WendlandC6Kernel` instead.
    """

    def __init__(self):
        super().__init__(
            (_WendlandC6Kernel(), DiracDeltaKernel(), _GaussianKernel(truncate=6.0))
        )
        self.size_in_fwhm = None  # initialized during Martini.__init__
        self._rescale = None  # initialized during Martini.__init__
        return

    def kernel(self, q):
        """
        Evaluate the kernel function of the WendlandC6 kernel.

        The WendlandC6 kernel is here defined as:

        .. math::
            W(q) = \\begin{cases}
            \\frac{1365}{64 \\pi} (1 - q)^8 (1 + 8q + 25q^2 + 32q^3)
            &{\\rm for}\\;0 \\leq q < 1\\\\
            0 &{\\rm for}\\;q \\geq 1
            \\end{cases}

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """
        return super().kernel(q)


class CubicSplineKernel(_AdaptiveKernel):
    """
    Implementation of the cubic spline (M4) kernel integral.

    The cubic spline is the 'classic' SPH kernel. The exact integral is usually
    too slow to be practical; the implementation here approximates the kernel
    amplitude as constant across the pixel, which converges to within 1% of
    the exact integral provided the SPH smoothing lengths are at least 1.16
    pixels in size.

    The cubic spline kernel is here defined as:

    .. math ::
        W(q) = \\frac{8}{\\pi}\\begin{cases}
        (1 - 6q^2(1 - \\frac{q}{2}))
        &{\\rm for}\\;0 \\leq q < \\frac{1}{2}\\\\
        2(1 - q)^3
        &{\\rm for}\\;\\frac{1}{2} \\leq q < 1\\\\
        0
        &{\\rm for}\\;q \\geq 1
        \\end{cases}

    This class falls back to the :class:`~martini.sph_kernels.DiracDeltaKernel` and
    :class:`~martini.sph_kernels.GaussianKernel` when the approximation used for the
    surface integral of the cubic spline kernel is not accurate to within better than 1%.
    To strictly use the cubic spline kernel, use
    :class:`~martini.sph_kernels._CubicSplineKernel` instead.
    """

    def __init__(self):
        super().__init__(
            (_CubicSplineKernel(), DiracDeltaKernel(), _GaussianKernel(truncate=6.0))
        )
        self.size_in_fwhm = None  # initialized during Martini.__init__
        self._rescale = None  # initialized during Martini.__init__
        return

    def kernel(self, q):
        """
        Evaluate the kernel function of the cubic spline kernel.

        The cubic spline kernel is here defined as:

        .. math ::
            W(q) = \\frac{8}{\\pi}\\begin{cases}
            (1 - 6q^2(1 - \\frac{q}{2}))
            &{\\rm for}\\;0 \\leq q < \\frac{1}{2}\\\\
            2(1 - q)^3
            &{\\rm for}\\;\\frac{1}{2} \\leq q < 1\\\\
            0
            &{\\rm for}\\;q \\geq 1
            \\end{cases}

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """
        return super().kernel(q)


class GaussianKernel(_AdaptiveKernel):
    """
    Implementation of a (truncated) Gaussian kernel integral.

    Calculates the kernel integral over a pixel. The 3 integrals (along :math:`dx`,
    :math:`dy`, :math:`dz`) are evaluated exactly, however the truncation is implemented
    approximately, erring on the side of integrating slightly further than
    the truncation radius.

    The Gaussian kernel is here defined as:

    .. math::
        W(q) = \\begin{cases}
        (\\sqrt{2\\pi}\\sigma)^{-3}
        \\exp\\left(-\\frac{1}{2}\\left(\\frac{q}{\\sigma}\\right)^2\\right)
        &{\\rm for}\\;0 \\leq q < t\\\\
        0 &{\\rm for}\\;q > t
        \\end{cases}

    with :math:`\\sigma=(2\\sqrt{2\\log(2)})^{-1}`, s.t. FWHM = 1, and
    :math:`t` being the truncation radius.

    This class falls back to the :class:`~martini.sph_kernels.DiracDeltaKernel` and
    :class:`~martini.sph_kernels.GaussianKernel` (with large truncation) when the
    approximation used for the surface integral of the Gaussian kernel is not accurate to
    within better than 1%. To strictly use the Gaussian kernel, use
    :class:`~martini.sph_kernels._GaussianKernel` instead.

    Parameters
    ----------
    truncate : float, optional
        Number of standard deviations at which to truncate kernel.
        Truncation radii <2 would lead to large errors and are not permitted.
        (Default: ``3``)
    """

    def __init__(self, truncate=3.0):
        super().__init__(
            (
                _GaussianKernel(truncate=truncate),
                DiracDeltaKernel(),
                _GaussianKernel(truncate=6.0),
            )
        )
        self.size_in_fwhm = None  # initialized during Martini.__init__
        self._rescale = None  # initialized during Martini.__init__
        return

    def kernel(self, q):
        """
        Evaluate the kernel function of the Gaussian kernel.

        The Gaussian kernel is here defined as:

        .. math::
            W(q) = \\begin{cases}
            (\\sqrt{2\\pi}\\sigma)^{-3}
            \\exp\\left(-\\frac{1}{2}\\left(\\frac{q}{\\sigma}\\right)^2\\right)
            &{\\rm for}\\;0 \\leq q < t\\\\
            0 &{\\rm for}\\;q > t
            \\end{cases}

        with :math:`\\sigma=(2\\sqrt{2\\log(2)})^{-1}`, s.t. FWHM = 1, and
        :math:`t` being the truncation radius.

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """
        return super().kernel(q)


class QuarticSplineKernel(_AdaptiveKernel):
    """
    Implementation of the quartic spline kernel integral.

    The quartic spline (M5) kernel is used in the SPHENIX scheme (e.g. in Colibre). The
    exact integral is usually too slow to be practical; the implementation here
    approximates the kernel amplitude as constant across the pixel, which converges to
    within 1% of the exact integral provided the SPH smoothing lengths are at least 1.2385
    pixels in size.

    The quartic spline kernel is here defined as:

    .. math ::
        W(q) = \\frac{15625}{512\\pi}\\begin{cases}
        (1 - q)^4 - 5(\\frac{3}{5} - q)^4 + 10(\\frac{1}{5}-q)^4
        &{\\rm for}\\;0 \\leq q < \\frac{1}{5}\\\\
        (1 - q)^4 - 5(\\frac{3}{5} - q)^4
        &{\\rm for}\\;\\frac{1}{5} \\leq q < \\frac{3}{5}\\\\
        (1 - q)^4
        &{\\rm for}\\;\\frac{3}{5} \\leq q < 1\\\\
        0
        &{\\rm for}\\;q \\geq 1
        \\end{cases}

    This class falls back to the :class:`~martini.sph_kernels.DiracDeltaKernel` and
    :class:`~martini.sph_kernels.GaussianKernel` when the approximation used for the
    surface integral of the quartic spline kernel is not accurate to within better
    than 1%. To strictly use the quartic spline kernel, use
    :class:`~martini.sph_kernels._QuarticSplineKernel` instead.
    """

    def __init__(self):
        super().__init__(
            (_QuarticSplineKernel(), DiracDeltaKernel(), _GaussianKernel(truncate=6.0))
        )
        self.size_in_fwhm = None  # initialized during Martini.__init__
        self._rescale = None  # initialized during Martini.__init__
        return

    def kernel(self, q):
        """
        Evaluate the kernel function of the quartic spline kernel.

        The quartic spline kernel is here defined as:

        .. math ::
            W(q) = \\frac{15625}{512\\pi}\\begin{cases}
            (1 - q)^4 - 5(\\frac{3}{5} - q)^4 + 10(\\frac{1}{5}-q)^4
            &{\\rm for}\\;0 \\leq q < \\frac{1}{5}\\\\
            (1 - q)^4 - 5(\\frac{3}{5} - q)^4
            &{\\rm for}\\;\\frac{1}{5} \\leq q < \\frac{3}{5}\\\\
            (1 - q)^4
            &{\\rm for}\\;\\frac{3}{5} \\leq q < 1\\\\
            0
            &{\\rm for}\\;q \\geq 1
            \\end{cases}

        Parameters
        ----------
        q : ~numpy.typing.ArrayLike
            Dimensionless distance parameter.

        Returns
        -------
        out : ~numpy.typing.ArrayLike
            Kernel value at positions ``q``.
        """
        return super().kernel(q)


class AdaptiveKernel(object):
    """
    Pre-configured adaptive kernels have been implemented in v2.0.3.

    See the SPH Kernels page in the documentation for details. You most likely
    want to use ``WendlandC2Kernel()``, ``WendlandC6Kernel()``, ``CubicSplineKernel()`` or
    ``QuarticSplineKernel()`` where you previously used ``AdaptiveKernel(...)``.

    Parameters
    ----------
    *args : any
        Any arguments, this class just raises an exception on attempted use.

    **kwargs : any
        Any keyword arguemnts; this class just raises an exception on attempted use.
    """

    def __init__(self, *args, **kwargs):
        raise NotImplementedError(
            "Pre-configured adaptive kernels have been implemented in v2.0.3, see "
            "the SPH Kernels page in the documentation for details. You most likely "
            "want to use WendlandC2Kernel(), WendlandC6Kernel(), CubicSplineKernel() or "
            "QuarticSplineKernel() where you previously used AdaptiveKernel(...)."
        )



"""
Provides classes for modelling the 21-cm spectral line emitted by a SPH particle.
"""

import numpy as np
import astropy.units as U
from astropy import constants as C
from scipy.special import erf
from abc import ABCMeta, abstractmethod


class _BaseSpectrum(metaclass=ABCMeta):
    """
    Abstract base class for implementions of spectral models to inherit from.

    Classes inheriting from :class:`~martini.spectral_models._BaseSpectrum` must implement
    two methods: :meth:`~martini.spectral_models._BaseSpectrum.half_width` and
    :meth:`~martini.spectral_models._BaseSpectrum.spectral_function`.

    :meth:`~martini.spectral_models._Base_spectrum.half_width` should define a
    characteristic width for the model, measured from the peak to the characteristic
    location. Note that particles whose spectra within +/- 4 half-widths of the peak do
    not intersect the data cube bandpass will be discarded to speed computation.

    :meth:`~martini.spectral_models._BaseSpectrum.spectral_function` should define the
    model spectrum. The spectrum should integrate to 1, the amplitude is handled
    separately.

    They may also override the method
    :meth:`~martini.spectral_models._BaseSpectrum.init_spectral_function_extra_data` to
    make information that depends on the :class:`~martini.sources.sph_source.SPHSource`
    (or derived class) or :class:`~martini.datacube.DataCube` properties available
    internally. This is required because the source object is not accessible at class
    initialization.

    Parameters
    ----------
    ncpu : int, optional
        Number of cpus to use for evaluation of particle spectra. Defaults to ``1`` if not
        provided. (Default: ``None``)

    spec_dtype : type, optional
        Data type of the arrays storing spectra of each particle, can be used to manage
        memory usage by adjusting precision.

    See Also
    --------
    martini.spectral_models.GaussianSpectrum
    martini.spectral_models.DiracDeltaSpectrum
    """

    def __init__(self, ncpu=None, spec_dtype=np.float64):
        self.ncpu = ncpu if ncpu is not None else 1
        self.spectral_function_extra_data = None
        self.spectra = None
        self.spec_dtype = spec_dtype
        return

    def init_spectra(self, source, datacube):
        """
        Pre-compute the spectrum of each particle.

        The spectral model defined in
        :meth:`~martini.spectral_models._BaseSpectrum.spectral_function` is evaluated
        using the channel edges from the :class:`~martini.datacube.DataCube` instance and
        the particle velocities of the :class:`~martini.sources.sph_source.SPHSource` (or
        derived class) instance provided.

        If the instance of this class was initialized with ``ncpu > 1`` then a
        process pool is created to distribute subsets of the calculation in
        parallel. To minimize overhead form serializing large amounts of
        data in :mod:`multiprocess` communications, each parallel process inherits the
        entire line-of-sight velocity array (cheap because of copy-on-write
        behaviour), then masks its copy to the subset to operate on.

        Parameters
        ----------
        source : ~martini.sources.sph_source.SPHSource
            Source object containing arrays of particle properties.

        datacube : ~martini.datacube.DataCube
            :class:`~martini.datacube.DataCube` object defining the observational
            parameters, including spectral channels.
        """

        self.channel_edges = datacube.velocity_channel_edges
        channel_widths = np.abs(np.diff(self.channel_edges).to(U.km * U.s**-1))
        self.vmids = source.skycoords.radial_velocity
        A = source.mHI_g * np.power(source.skycoords.distance.to(U.Mpc), -2)
        MHI_Jy = (
            U.Msun * U.Mpc**-2 * (U.km * U.s**-1) ** -1,
            U.Jy,
            lambda x: (1 / 2.36e5) * x,
            lambda x: 2.36e5 * x,
        )
        if self.ncpu == 1:
            raw_spectra = self.evaluate_spectra(source, datacube)
        else:
            from multiprocess.pool import Pool

            with Pool(processes=self.ncpu) as pool:
                raw_spectra = np.vstack(
                    pool.map(
                        lambda mask: self.evaluate_spectra(source, datacube, mask=mask),
                        [
                            (
                                np.s_[
                                    icpu
                                    * len(self.vmids)
                                    // self.ncpu : (icpu + 1)
                                    * len(self.vmids)
                                    // self.ncpu
                                ]
                                if icpu is not None
                                else np.s_[...]
                            )
                            for icpu in range(self.ncpu)
                        ],
                    )
                )
        self.spectra = (
            A.astype(self.spec_dtype)[..., np.newaxis]
            * raw_spectra
            / channel_widths.astype(self.spec_dtype)
        ).to(U.Jy, equivalencies=[MHI_Jy])

        return

    def evaluate_spectra(self, source, datacube, mask=np.s_[...]):
        """
        The main portion of the calculation of the spectra.

        Separated into this function so that it can be called by a parallel
        process pool. Initializes additional particle properties by calling
        :meth:`~martini.spectral_models._BaseSpectrum.init_spectral_function_extra_data`
        which then becomes accessible via
        :attr:`~martini.spectral_models._BaseSpectrum.spectral_function_extra_data`.

        Parameters
        ----------
        source : ~martini.sources.sph_source.SPHSource
            Source object containing arrays of particle properties.

        datacube : ~martini.datacube.DataCube
            :class:`~martini.datacube.DataCube` object defining the observational
            parameters, including spectral channels.

        mask : slice, optional
            Slice defining the subset of particles to operate on.
            (Default: ``np.s_[...]``)
        """
        vmids = self.vmids[mask]
        self.init_spectral_function_extra_data(source, datacube, mask=mask)
        if all(np.diff(self.channel_edges) > 0):
            lower_edges_slice = np.s_[:-1]
            upper_edges_slice = np.s_[1:]
        elif all(np.diff(self.channel_edges) < 0):
            lower_edges_slice = np.s_[1:]
            upper_edges_slice = np.s_[:-1]
        else:
            raise ValueError("Channel edges are not monotonic sequence.")
        return self.spectral_function(
            (
                np.tile(
                    self.channel_edges.to_value(self.channel_edges.unit)[
                        lower_edges_slice
                    ],
                    vmids.shape + (1,),
                )
                * self.channel_edges.unit
            ).astype(self.spec_dtype),
            (
                np.tile(
                    self.channel_edges.to_value(self.channel_edges.unit)[
                        upper_edges_slice
                    ],
                    vmids.shape + (1,),
                )
                * self.channel_edges.unit
            ).astype(self.spec_dtype),
            (
                np.tile(
                    vmids.to_value(vmids.unit),
                    np.shape(self.channel_edges[:-1]) + (1,) * vmids.ndim,
                ).T
                * vmids.unit
            ).astype(self.spec_dtype),
        ).to_value(U.dimensionless_unscaled)

    @abstractmethod
    def half_width(self, source):
        """
        Abstract method; calculate the half-width of the spectrum, either globally or
        per-particle.

        Parameters
        ----------
        source : ~martini.sources.sph_source.SPHSource
            The source object will be provided to allow access to particle
            properties.
        """
        pass

    @abstractmethod
    def spectral_function(self, a, b, vmids):
        """
        Abstract method; implementation of the spectral model.

        Should calculate the flux in each spectral channel, calculation should
        be vectorized (with :mod:`numpy`).

        Parameters
        ----------
        a : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Lower spectral channel edge(s).

        b : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Upper spectral channel edge(s).

        vmids : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Particle velocities along the line of sight.

        See Also
        --------
        martini.spectral_models._BaseSpectrum.init_spectral_function_extra_data
        """

        pass

    def init_spectral_function_extra_data(self, source, datacube, mask=np.s_[...]):
        """
        Initialize extra data needed by spectral function. Default is no extra data.

        Derived classes should override this function, if needed, to populate the dict
        with any information from the source that is required by the
        :meth:`~martini.spectral_models._BaseSpectrum.spectral_function`,
        then call ``super().init_spectral_function_extra_data``.

        Parameters
        ----------
        source : ~martini.sources.sph_source.SPHSource
            Source object, making particle properties available.

        datacube : ~martini.datacube.DataCube
            :class:`~martini.datacube.DataCube` object defining the observational
            parameters, including spectral channels.

        mask : slice, optional
            Slice defining the subset of particles to operate on.
            (Default: ``np.s_[...]``)

        See Also
        --------
        martini.spectral_models.GaussianSpectrum.init_spectral_function_extra_data
        """
        if self.spectral_function_extra_data is None:
            self.spectral_function_extra_data = dict()
        self.spectral_function_extra_data = {
            k: np.tile(
                v[mask] if not v.isscalar else v,
                np.shape(datacube.channel_edges[:-1])
                + (1,) * source.skycoords.radial_velocity.ndim,
            )
            .astype(self.spec_dtype)
            .T
            for k, v in self.spectral_function_extra_data.items()
        }
        return


class GaussianSpectrum(_BaseSpectrum):
    """
    Class implementing a Gaussian model for the spectrum of the HI line.

    The line is modelled as a Gaussian of either fixed width, or of width
    scaling with the particle temperature as :math:`\\sqrt{k_B T / m_p}`, centered
    at the particle velocity.

    Parameters
    ----------
    sigma : ~astropy.units.Quantity or str, optional
        :class:`~astropy.units.Quantity`, with dimensions of velocity, or string
        ``"thermal"``.
        Width of the Gaussian modelling the line (constant for all particles),
        or specify ``"thermal"`` for width equal to :math:`\\sqrt{k_B T / m_p}` where
        :math:`k_B` is Boltzmann's constant, :math:`T` is the particle temperature and
        :math:`m_p` is the particle mass. (Default: ``7 U.km * U.s**-1``)

    ncpu : int, optional
        Number of cpus to use for evaluation of particle spectra. Defaults to ``1`` if not
        provided. (Default: ``None``)

    spec_dtype : type, optional
        Data type of the arrays storing spectra of each particle, can be used to manage
        memory usage by adjusting precision.

    See Also
    --------
    martini.spectral_models._BaseSpectrum
    martini.spectral_models.DiracDeltaSpectrum
    """

    def __init__(self, sigma=7.0 * U.km * U.s**-1, ncpu=None, spec_dtype=np.float64):
        self.sigma_mode = sigma
        super().__init__(ncpu=ncpu, spec_dtype=spec_dtype)

        return

    def spectral_function(self, a, b, vmids):
        """
        Evaluate a Gaussian integral in a channel. Requires sigma to be available from
        :attr:`~martini.spectral_models.GaussianSpectrum.spectral_function_extra_data`.

        Parameters
        ----------
        a : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Lower spectral channel edge(s).

        b : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Upper spectral channel edge(s).

        vmids : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Particle velocities along the line of sight.

        Returns
        -------
        out : ~astropy.units.Quantity
            The evaluated spectral model (dimensionless).
        """

        assert self.spectral_function_extra_data is not None
        sigma = self.spectral_function_extra_data["sigma"]

        return self.spec_dtype(0.5) * (
            erf((b - vmids) / (np.sqrt(self.spec_dtype(2.0)) * sigma))
            - erf((a - vmids) / (np.sqrt(self.spec_dtype(2.0)) * sigma))
        )

    def init_spectral_function_extra_data(self, source, datacube, mask=np.s_[...]):
        """
        Helper function to expose particle velocity dispersions to
        :meth:`~martini.spectral_models.GaussianSpectrum.spectral_function`.

        Parameters
        ----------
        source : ~martini.sources.sph_source.SPHSource
            Source object.

        datacube : ~martini.datacube.DataCube
            :class:`~martini.datacube.DataCube` object.

        mask : slice, optional
            Slice defining the subset of particles to operate on.
            (Default: ``np.s_[...]``)
        """

        self.spectral_function_extra_data = dict(sigma=self.half_width(source))
        super().init_spectral_function_extra_data(source, datacube, mask=mask)
        return

    def half_width(self, source):
        """
        Calculate 1D velocity dispersions from particle temperatures, or return
        constant.

        Parameters
        ----------
        source : ~martini.sources.sph_source.SPHSource
            Source object, making particle properties available.

        Returns
        -------
        out : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Velocity dispersion (constant, or per particle).
        """

        if self.sigma_mode == "thermal":
            # 3D velocity dispersion of an ideal gas is sqrt(3 * kB * T / mp)
            # So 1D velocity dispersion is sqrt(kB * T / mp)
            return np.sqrt(C.k_B * source.T_g / C.m_p).to(U.km * U.s**-1)
        else:
            return self.sigma_mode


class DiracDeltaSpectrum(_BaseSpectrum):
    """
    Class implemeting a Dirac-delta model for the spectrum of the HI line.

    The line is modelled as a Dirac-delta function, centered at the particle
    velocity.

    Parameters
    ----------
    ncpu : int, optional
        Number of cpus to use for evaluation of particle spectra. Defaults to ``1`` if not
        provided. (Default: ``None``)

    spec_dtype : type, optional
        Data type of the arrays storing spectra of each particle, can be used to manage
        memory usage by adjusting precision.
    """

    def __init__(self, ncpu=None, spec_dtype=np.float64):
        super().__init__(ncpu=ncpu, spec_dtype=spec_dtype)
        return

    def spectral_function(self, a, b, vmids):
        """
        Evaluate a Dirac-delta function in a channel.

        Parameters
        ----------
        a : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Lower spectral channel edge(s).

        b : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Upper spectral channel edge(s).

        vmids : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Particle velocities along the line of sight.

        Returns
        -------
        out : ~astropy.units.Quantity
            The evaluated spectral model (dimensionless).
        """

        return np.heaviside(vmids - a, 1.0) * np.heaviside(b - vmids, 0.0)

    def half_width(self, source):
        """
        Dirac-delta function has 0 width.

        Parameters
        ----------
        source : ~martini.sources.sph_source.SPHSource
            Source object, making particle properties available.

        Returns
        -------
        out : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity`, with dimensions of velocity.
            Velocity dispersion of ``0 * U.km * U.s**-1``.
        """

        return 0 * U.km * U.s**-1
